/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.fwutil;

import net.java.dev.hvlayout.*;

import java.awt.BorderLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
//import java.awt.Font;
//import java.awt.Insets;

import javax.swing.JFrame;
import javax.swing.UIManager;
import javax.swing.WindowConstants;

/**
 * A (singleton) window showing the log-messages from Log.
 * Use setProgrammaticClose() to prevent the applicaiton from exiting
 * when this log-window is closed.
 * @author Fred
 *
 */
public class LogUI extends JFrame implements ParentWindow {

	private static final long serialVersionUID = 1L;

	protected static final ArrayList<MFrame> childWindows = new ArrayList<MFrame>();

	private static LogUI thisInstance = new LogUI();
	private static HVTextArea text;
	private static WindowAdapter closer;

	private LogUI() {
		
		super(Settings.appName + " " + Settings.appVersion + " Log");
		try {
			UIManager.setLookAndFeel(
					UIManager.getSystemLookAndFeelClassName());
		} catch (Exception unused) { }

		closer = new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				Startup.close(0);
			}
		};
		addWindowListener(closer);

		text = FormUI.getTextArea(Util.emptyString, true);
		text.setLineWrap(true);
		this.getContentPane().add(HVProps.getScroller(text, 4 * HVProps.defaultLineWidth, 20, false), BorderLayout.CENTER);
		this.pack();
		WindowDimension.positionAtMiddleOfScreen(this);
	}
	
	/** Returns the only instance of this LogUI. */
	public static LogUI getInstance() {
		
		return thisInstance;
	}
	
	public static HVTextArea getTextArea() {
		return text;
	}
	
	public static void setCloseListener(WindowAdapter customCloser) {
		thisInstance.removeWindowListener(closer);
		thisInstance.addWindowListener(customCloser);
	}
	
	/** If called, prevents the program from exiting when this LogUI window is closed. */
	public static void setProgrammaticClose() {
		
		thisInstance.removeWindowListener(closer);
		thisInstance.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);
	}
	
	/** Hides the log-window. */
	public static boolean makeInvisible() {

		if (!thisInstance.isVisible())
			return false;
		thisInstance.setVisible(false);
		return true;
	}
	
	/** Shows the log-window. */
	public static boolean makeVisible() {

		if (thisInstance.isVisible()) {
			thisInstance.toFront();
			return false;
		}
		thisInstance.setVisible(true);
		thisInstance.toFront();
		return true;
	}
	
	/**
	 * Shows the log-message in the log-window.
	 * @param message The log-message to show.
	 */
	public static void showMessage(String message) {
		
		text.append(message + Util.newline);
		if (thisInstance.isVisible()) { 
			text.setCaretPosition(text.getDocument().getLength());
/*			// Following paintImmediatly is not really necessary
 			// (the message has already been written to disk)
 			// and has quite some performance impact.
 			// Other components do need this tweak sometimes, 
 			// so leave it here for lookup/howto. 
			if (Settings.appDebug)
				text.paintImmediately(text.getBounds());
			else
*/				text.repaint();
		}
	}
	
	/** Reduces the shown log-text to 10% of it's size.
	 * This is to prevent delays in showMessage(). 
	 * @param maxLogFileSize Used in the calculation to determine what 10% is.
	 */ 
	public static void chopText(int maxLogFileSize) {

		String chopLog = text.getText();
		int chopLogSize = chopLog.length() / 10;
		if (chopLogSize > (maxLogFileSize / 10))
			chopLogSize = maxLogFileSize / 10;
		chopLog = chopLog.substring(chopLog.length() - chopLogSize,  chopLog.length());
		text.setText(chopLog);
		text.setCaretPosition(text.getDocument().getLength());
		text.repaint();
	}
	
	public void addChildWindow(MFrame childWindow) {
		
		if (!childWindows.contains(childWindow)) {
			childWindows.add(childWindow);
			Log.log("Added window " + childWindow.getTitle());
		}
	}
	
	public boolean removeChildWindow(MFrame childWindow) {
		
		boolean removed = childWindows.remove(childWindow);
		if (removed) {
			Log.log("Removed window " + childWindow.getTitle() + ", " + openChildWindows() + " windows still registered.");
		}
		return childWindows.remove(childWindow);
	}
	
	protected void closeChildWindows() {
		
		while (childWindows.size() > 0)
			childWindows.get(0).dispose();
	}
	
	public static int openChildWindows() {
		return childWindows.size();
	}
	
	@SuppressWarnings("unchecked")
	public static ArrayList<MFrame> getChildWindows() {
		return (ArrayList<MFrame>) childWindows.clone();
	}


	/** Closes the log-window.
	 * If StartUp.restart is false, also disposes this log-window.
	 */
	public static void close() {

		// Log is usually closed as last, because of that the following will not work:
		// WindowDimension.setWindowDimension("LogFrame", thisInstance);
		// The applyWindowDimension is called in Startup.boot()
		// The setWindowDimension is called in Startup.close()
		thisInstance.closeChildWindows();
		thisInstance.setVisible(false);
		if (!Startup.restart)
				thisInstance.dispose();
	}
}
