/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.fwutil;

import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.XMLEncoder;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.WindowConstants;

import net.java.dev.hvlayout.*;

/**
 * An example on how to use the capabilities of FWUtil in a desktop-java application.
 * This example shows a main-class that initializes the environment
 * and starts the main window 'MainFrame'. Read the comments in the source
 * to get an overview of what is going on and how FWUtil is used.
 *
 */
public class MainAppExample implements Restartable {
	
	private static JButton bShowLog;
	private static JButton bRestart;
	private static JButton bSelectFont;
	private static JButton bExit;
	private static JCheckBox bClearSettings;
	
	// Needed in case of restart - MainFrame passes this Restartable on to
	// StartUp.close(Restartable);
	private static MainAppExample thisInstance = new MainAppExample();
	
	private MainAppExample() {
		super();
		// Needs to be done only once.
		registerXMLEncoderDelegates();
		thisInstance = this;
	}
	
	public static MainAppExample getInstance() {
		return thisInstance;
	}

	public static void main(String[] args) {

		// Since we do not use a separate application data folder
		// (see the StartUp.boot call further on)
		// set a non-default settings file so that we do not overwrite or
		// load a settings file appropriate for another application.
		Settings.defaultSettingsFile = "MainAppExamplesettings.xml";
		// Load the labels from a file specific for this application.
		// This can also be done from the command line using the parameter:
		// -labels=labelsfile.txt
		// The label-file must use UTF-8 encoding.
		Labels.labelFile = Util.confDir + Util.fileSeparator + "MainAppExampleLabels.txt";
		// StartUp - load labels and settings, etc.
		// By specifying null as data-dir, the java-defined temp folder will be used.
		if (!Startup.boot(args, "MainAppExample", "0.1", null)) {
			Log.log("Startup failed, exiting.");
			UIMessages.showErrorMessage(LogUI.getInstance(), "StartUp error", "Error during startup, please examine log.\nClosing " + Settings.appName);
			Log.close();
			System.exit(1);
		}
		// StartUp went OK, start the main Window.
		Log.log(Labels.get("MainFrameStart"));
		// Use SwingUtilities.invokeLater to prevent rare occasions where
		// new JFrames simply do not appear.
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				thisInstance.new MainFrame("MainAppExample");
			}
		});
	}

	/**
	 * This example does not use SortKey but this function shows what
	 * to do if it is used. Simply registering the persistence delegate
	 * is enough to allow SortKey values to be written to and 
	 * read from the Settings.
	 *
	 */
	public static void registerXMLEncoderDelegates() {

		// Following two statements allow SortKery values to be stored in Settings.
		XMLEncoder encoder = new XMLEncoder(System.out);
		encoder.setPersistenceDelegate(javax.swing.RowSorter.SortKey.class, new SortKeyPersistenceDelegate());
	}

	/** The restart method as defined by the Restartable interface. 
	 * Simply calls main again.
	 */
	public void restart(String[] args) {
		main(args);
	}
	
	/**
	 * The main window of the example application.
	 * Note how the closing operation is set during initialization.
	 * This window does not use MFrame or ParentWindow,
	 * for an example use of those two see net.java.dev.fwutil.filebrowser.FileBrowser.
	 *
	 */
	@SuppressWarnings("serial")
	class MainFrame extends JFrame implements ActionListener {
		
		public MainFrame(String title) {

			super(Settings.appName + " " + Settings.appVersion);
			// Prevent closing the log from closing the application.
			LogUI.setProgrammaticClose();
			// Set a custom closing operation: close().
			setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
			bExit = FormUI.getWideButton(Labels.get("Exit"));
			bExit.addActionListener(this);
			addWindowListener(new WindowCloseAdapter(bExit, this));

			Container mainPanel = this.getContentPane();
			mainPanel.setLayout(new VLayout());
			registerButtons(mainPanel);
			pack();
			// Apply or store the window place and size on screen.
			// Always do this after pack().
			WindowDimension.applyWindowDimension(this.getTitle(), this);
			// Keep log-window open only in debug-mode.
			if (!Settings.appDebug)
				Log.makeInvisible();
			setVisible(true);
		}
		
		/**
		 * This method closes the application after storing all appropriate
		 * information.
		 *
		 */
		public void close() {
			if (!UIMessages.askYesNo(this, Labels.get("Exit"), Labels.get("ReallyExitApp")))
				return;
			// Store the window place and size on screen in the Settings.
			WindowDimension.setWindowDimension(this.getTitle(), this);
			setVisible(false);
			dispose();
			// Try to delete the settings file when requested.
			if (bClearSettings.isSelected()) {
				if (!Util.deleteFile(Settings.defaultSettingsFile))
					System.out.println("Failed to delete settings file: " +
							Settings.defaultSettingsFile);
				else
					System.out.println("Settings file deleted: " +
							Settings.defaultSettingsFile);
				// By default, StartUp.close() will write the Settings in a
				// setting-file. 
				Startup.doNotWriteSettings = true;
			}
			// Close and clear all FWUtil classes. Settings are stored if
			// StartUp.doNotWriteSettings is false.
			Startup.close(MainAppExample.getInstance());
		}
		
		/**
		 * Registers and places the action-buttons on a panel.
		 * Note that FontSelector registers the selected font (size)
		 * in Settings. In turn, StartUp.boot calls Settings.setDefaultFont()
		 * which applies the font as selected in FontSelector.
		 * @param vbox The panel containing the buttons.
		 */
		public void registerButtons(Container vbox) {
			
			bShowLog = FormUI.getWideButton(Labels.get("ShowLog"));
			bShowLog.addActionListener(this);
			vbox.add(bShowLog);
			bSelectFont = FormUI.getWideButton(Labels.get("SelectFont"));
			bSelectFont.addActionListener(this);
			vbox.add(bSelectFont);
			bRestart = FormUI.getWideButton(Labels.get("Restart"));
			bRestart.addActionListener(this);
			vbox.add(bRestart);
			vbox.add(bExit);
			bClearSettings = FormUI.getCheckBox(Labels.get("DeleteSettings"), HVProps.defaultLineWidth * 2);
			// bClearSettings is used in the close() operation.
			vbox.add(bClearSettings);
		}

		/**
		 * Performs the actions for the action-buttons.
		 */
		public void actionPerformed(ActionEvent e) {
			
			if (e.getSource() == bShowLog) {
				LogUI.makeVisible();
			} else if (e.getSource() == bRestart) {
				Startup.restart = true;
				close();
			} else if (e.getSource() == bSelectFont) {
				FontSelector.show(this);
			} else if (e.getSource() == bExit) {
				close();
			}
		}
	}
}
