/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.fwutil;

import net.java.dev.hvlayout.*;

import java.awt.Font;
import java.beans.XMLEncoder;
import java.util.Hashtable;

/**
 * A singleton class that contains a table with keys and values (settings).
 * This class also contains variables that are used throughout
 * the fwutil package (e.g. appDebug).
 * <br> The settings can be stored and loaded. This is done
 * by StartUp.boot() and StartUp.close(). 
 * Reading and writing is done using the
 * XMLEncoder which can deal with anything that conforms to
 * the Java-beans specification (see WindowDimension for example)
 * and by default all java.util stuff and all java primitives.
 * <br> By registering a persistence delegate, other classes
 * can be stored as well. See for example SortKeyPersistenceDelegate
 * at the static part of this class.
 * <p>
 * This class is not type-safe. Important "global" parameters 
 * can be read and stored using this class, but a separate class
 * should provide type-safety and default values for these parameters. 
 * @author Fred
 *
 */
public final class Settings {

	static {
		// Following two statements allow SortKery values to be stored and retrieved in Settings.
		// Only needs to be done once.
		XMLEncoder encoder = new XMLEncoder(System.out);
		encoder.setPersistenceDelegate(javax.swing.RowSorter.SortKey.class, new SortKeyPersistenceDelegate());
	}

	@SuppressWarnings("unused")
	private static Settings thisInstance = new Settings();
	@SuppressWarnings("unchecked")
	private static Hashtable settings;

	/** The name of the application. See also StartUp.boot() */
	public static String appName = "appName";
	/** The version of the application. See also StartUp.boot()  */
	public static String appVersion = "appVersion";
	/** The name of the file with the settings. */
	public static String defaultSettingsFile = "settings.xml";
	/** Indicates of debug mode is on or off. */
	public static boolean appDebug = false;
	/** The data directory of the application (containing the settings file
	 * and the log-files
	 */
	public static String appDataDir = null;
	/** The root directory of the software of the application. */
	public static String appHomeDir = null;
	/** Indicates if the application has been started before.
	 Set to true when a Settings file is NOT found. */
	public static boolean firstStart = false;
	/** The default font size set by the user (see FontSelector). 
	 * When the settings file is loaded and this key is detected, 
	 * the UI is updated to use the indicated font size.
	 */
	public static final String defaultFontSize = "DefaultFontSize";
	/** The default font set by the user (see FontSelector). 
	 * When the settings file is loaded and this key is detected, 
	 * the UI is updated to use the indicated font.
	 */
	public static final String defaultFontName = "DefaultFontName"; 
	
	@SuppressWarnings("unchecked")
	private Settings() {
		
		super();
		settings = new Hashtable();
	}
	
	/**
	 * Returns the Hashtable containing all the data.
	 */
	@SuppressWarnings("unchecked")
	public static Hashtable getSettings() {
		
		return settings;
	}
	
	/**
	 * Clears all settings from memory.
	 *
	 */
	public static void clearSettings() {
		
		settings.clear();
	}
	
	/**
	 * Sets newSettings as the source for all data.
	 */

	@SuppressWarnings("unchecked")
	public static void setSettings(Hashtable newSettings) {
		
		if (newSettings != null) {
			settings = newSettings;
//			putInitSettings();
		}
	}
	
	/**
	 * Returns the value associated with the key.
	 */
	public static Object get(Object key) {
		
		return settings.get(key);
	}
	
	@SuppressWarnings("unchecked")
	/**
	 * Stores the key and value in the settings.
	 */
	public static Object put(Object key, Object value) {
		
		return settings.put(key, value);
	}
	
	/**
	 * Removes the key and associated value from memory.
	 */
	public static Object remove(Object key) {
		
		return settings.remove(key);
	}
	
	/**
	 * Reads in settings data from a file.
	 * @return True if file was read succesfully.
	 */
	@SuppressWarnings("unchecked")
	public static boolean readSettings(String fileName) {
		
		Object o = Util.readObject(fileName);
		if (o == null)
			return false;
		if (o instanceof Hashtable) {
			settings = (Hashtable)o;
			return true;
		}
		return false;
	}

	/**
	 * Called by StartUp.boot() to update the UI with the user
	 * selected font (size). See also FontSelector.
	 */
	public static void setDefaultFont() {
		
		if (Settings.get(Settings.defaultFontSize) != null &&
				Settings.get(Settings.defaultFontName) == null) {
			int fsize = ((Number)Settings.get(Settings.defaultFontSize)).intValue();
			HVProps.setDefaultUIFontSize(fsize);
		}
		if (Settings.get(Settings.defaultFontName) != null) {
			int fsize = HVProps.getDefaultFontSize();
			if (Settings.get(Settings.defaultFontSize) != null)
				fsize = ((Number)Settings.get(Settings.defaultFontSize)).intValue();
			Font newFont = new Font(Settings.get(Settings.defaultFontName).toString(), Font.PLAIN, fsize);
			HVProps.setDefaultUIFont(newFont);
		}
	}
	
	/**
	 * Stores the settings in memory to file.
	 * @return False if writing to file failed.
	 */
	public static boolean writeSettings(String fileName) {
		
		return Util.writeObject(fileName, settings);
	}
}
