/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.fwutil;

/**
 * StringLine contains the FileStringIterator (FSI) object and a
 * function that depends on it: getNextLine(sfi).
 * <br> The FSI object and getNextLine(sfi) serve as a building block
 * for Util functions like bigStringToParameters(), getKeyValue()
 * and continueReadMultiLineValue().
 *
 */
public final class StringLine {
	
	/** The 'go to next line' character used by Windows */
	public static final char carriageReturn = '\r';
	/** The 'end of line' character used by Windows and Linux */ 
	public static final char lineFeed = '\n';

	private static StringLine thisInstance = new StringLine();
	
	private StringLine() {
		super();
	}
	
	/**
	 * A data-object that contains information about where reading the
	 * 'bigString' is at, use in combination with getNextLine(fsi).
	 */
	public class FileStringIterator {
		
		/** The string to chop into lines. */
		public final String bigString;
		/** The position to start fetching the next line from. */
		public int startFrom;
		/** The line-number currently ready to read, only works if startFrom was 0 upon instantiation of the FSI. */
		public int lineCount;
		
		public FileStringIterator() {
			
			this(null, 0);
		}
		
		public FileStringIterator(final String bigString, int startFrom) {

			super();
			this.bigString = bigString;
			this.startFrom = startFrom;
			this.lineCount = -1;
			if (startFrom == 0)
				this.lineCount = 1;
		}
	}
	
	/**
	 * Convenience method to get a new FSI.
	 * @param bigString The string to read into lines.
	 * @param startFrom The starting position to start reading, set to 0 to enable line-count.
	 * @return A new FileStringIterator ready for use by getNextLine(fsi).
	 */
	public static FileStringIterator getNewFSI(String bigString, int startFrom) {
		return thisInstance.new FileStringIterator(bigString, startFrom);
	}

	/**
	 * Convenience method to get a new FSI. bigString and startFrom need to be set by the programmer.
	 * @return The new FileStringIterator.
	 */
	public static FileStringIterator getNewFSI() {
		return thisInstance.new FileStringIterator();
	}
	
	/**
	 * Fetches the next line from the bigString in the FSI and updates
	 * startFrom and lineCount in the FSI as needed so that upon return
	 * this function is ready to fetch the next line from the FSI.
	 * <br> @see Util.bigStringToParameters depends on this function.
	 * @param sfi
	 * @return The next line.
	 */
	public static String getNextLine(FileStringIterator sfi) {
		
		if (sfi.startFrom >= sfi.bigString.length())
			return null;
		String line = null;
		char startChar = sfi.bigString.charAt(sfi.startFrom);
		
		while (sfi.startFrom < sfi.bigString.length() &&
				((startChar == lineFeed) || startChar == carriageReturn)) {
			if (startChar == lineFeed)
				sfi.lineCount++;
			sfi.startFrom++;
			if (sfi.startFrom < sfi.bigString.length())
				startChar = sfi.bigString.charAt(sfi.startFrom);
		}
		if (sfi.startFrom >= sfi.bigString.length())
			return null;
		int endIndex = sfi.bigString.indexOf(lineFeed, sfi.startFrom);
		if (endIndex == -1) {
			// reached end of string
			line = sfi.bigString.substring(sfi.startFrom, sfi.bigString.length());
			sfi.startFrom = sfi.bigString.length();
			return line;
		}
		// The DOS/Windows newline is \r\n
		if (sfi.bigString.charAt(endIndex-1) == carriageReturn)
			endIndex--;
		line = sfi.bigString.substring(sfi.startFrom, endIndex);
		sfi.startFrom = endIndex;
		return line;
	}
}