/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.fwutil;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PushbackInputStream;
import java.io.Reader;
import java.nio.charset.Charset;

/**
 * Generic unicode (UTF) textreader, which will use BOM mark to identify the
 * encoding to be used. Support for UTF32 has not been implemented yet.
 * <pre>
 * This is a modified version of the original file from 
 * <a href="http://www.velocityreviews.com/forums/t123963-textstreamreader-with-transparent-unicode-bom-support.html">UTF text stream reader</a>
 * which fixes a bug discussed at
 * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4508058"> Sun's bug-database </a>.
 * 
 * Original pseudocode : Thomas Weidenfeller 
 * Implementation tweaked: Aki Nieminen
 * 
 * http://www.unicode.org/unicode/faq/utf_bom.html 
 * BOMs: 
 * 00 00 FE FF = UTF-32, big-endian 
 * FF FE 00 00 = UTF-32, little-endian 
 * FE FF = UTF-16, big-endian
 * FF FE = UTF-16, little-endian 
 * EF BB BF = UTF-8
 * 
 * Tip for testing: Windows Notepad Unicode format = UTF-16LE
 * 
 * The String constants for identifying UTF encoding formats are taken from 
 * <a href="http://java.sun.com/j2se/1.4.2/docs/guide/intl/encoding.doc.html">encoding api doc</a>.
 * </pre>
 */

public class UnicodeReader extends Reader {
	
	public static final int BOM_SIZE = 4;
	public static final String defaultCharset = Charset.defaultCharset().name();
// Defaults copied form http://java.sun.com/j2se/1.4.2/docs/guide/intl/encoding.doc.html
	public static final String UTF8 = "UTF8";
	public static final String UTF8nio = "UTF-8";
	public static final String UTF16 = "UTF-16";
	public static final String UTF16BE = "UnicodeBigUnmarked";
	public static final String UTF16LE = "UnicodeLittleUnmarked";
	public static final String UTF16Default = UTF16BE;

	private PushbackInputStream internalIn;
	private InputStreamReader internalIn2 = null;
	private String defaultEnc;
	private String usedEncoding;
	private Boolean hasBom;
	

	/**
	 * Reads the first bytes of the inputstream to indetify the UTF Encoding to use. 
	 * If no UTF BOM could be found, the read bytes are unread and the default encoding specified by defaultEnc
	 * is used. 
	 */	
	public UnicodeReader(InputStream in, String defaultEnc) throws IOException {
		super();
		internalIn = new PushbackInputStream(in, BOM_SIZE);
		this.defaultEnc = defaultEnc;
		hasBom = false;
		init();
	}
	
	/**
	 * Opens a file and attempts to read UTF BOM-information to identify the UTF Encoding to use.
	 * If no BOM information is found the encoding specified by defaultEnc is used.
	 * @param defaultEnc
	 * The default encoding to use if no BOM-information is available to determine the UTF Encoding to use.
	 */
	public UnicodeReader(String fileName, String defaultEnc) throws IOException  {
 		this(new FileInputStream(fileName), defaultEnc);
	}
	
	/**
	 * Opens a file and attempts to read UTF BOM-information to identify the UTF Encoding to use.
	 * If no BOM information is found the system default Charset Encoding is used.
	 */
	public UnicodeReader(String fileName) throws IOException  {
 		this(new FileInputStream(fileName), defaultCharset);
	}

	/**
	 * Read-ahead four bytes and check for BOM marks. Extra bytes are
	 * unread back to the stream, only BOM bytes are skipped.
	 * UTF32 encoding will throw an "Unsupported encoding" exception.
	 */
	protected void init() throws IOException {
		if (internalIn2 != null) return;
		
		String encoding = null;
		byte bom[] = new byte[BOM_SIZE];
		int bomBytesRead, unread;
		
		bomBytesRead = internalIn.read(bom);
		hasBom = true;
		if ( (bom[0] == (byte)0xEF) && (bom[1] == (byte)0xBB) &&
				(bom[2] == (byte)0xBF) ) {
			encoding = "UTF-8";
			unread = bomBytesRead - 3;
		} else if ( (bom[0] == (byte)0xFE) && (bom[1] == (byte)0xFF) ) {
			encoding = "UTF-16BE";
			unread = bomBytesRead - 2;
		} else if ( (bom[0] == (byte)0xFF) && (bom[1] == (byte)0xFE) ) {
			encoding = "UTF-16LE";
			unread = bomBytesRead - 2;
		} else if ( (bom[0] == (byte)0x00) && (bom[1] == (byte)0x00) &&
				(bom[2] == (byte)0xFE) && (bom[3] == (byte)0xFF)) {
			encoding = "UTF-32BE";
			unread = bomBytesRead - 4;
		} else if ( (bom[0] == (byte)0xFF) && (bom[1] == (byte)0xFE) &&
				(bom[2] == (byte)0x00) && (bom[3] == (byte)0x00)) {
			encoding = "UTF-32LE";
			unread = bomBytesRead - 4;
		} else {
//			Unicode BOM mark not found, unread all bytes
			hasBom = false;
			encoding = defaultEnc;
			unread = bomBytesRead;
		}
//System.out.println("read=" + bomBytesRead + ", unread=" + unread + ", enc=" + encoding);
		if (unread > 0) internalIn.unread(bom, (bomBytesRead - unread), unread);
		else if (unread < -1) internalIn.unread(bom, 0, 0);
		if (encoding.equals("UTF-8"))
			encoding = UTF8;
		if (encoding.equals("UTF-16BE"))
			encoding = UTF16BE;
		if (encoding.equals("UTF-16LE"))
			encoding = UTF16LE;
		usedEncoding = encoding;
		if (encoding.equals("UTF-32BE"))
			throw new IOException("Unsupported encoding: UTF-32BE");
		if (encoding.equals("UTF-32LE"))
			throw new IOException("Unsupported encoding: UTF-32LE");
		internalIn2 = new InputStreamReader(internalIn, usedEncoding);
	}
	
	public void close() throws IOException {
//		init();
		internalIn2.close();
	}
	
	public int read(char[] cbuf, int off, int len) throws IOException {
//		init();
		return internalIn2.read(cbuf, off, len);
	}

	/**
	 * Returns the default encoding as set by the constructor or by the system.
	 */
	public String getDefaultEncoding() {
		return defaultEnc;
	}
	
	/**
	 * Returns the encoding used to read the text file. Can return null when no file was opened.
	 */
	public String getEncoding() {
		if (internalIn2 == null) return null;
		return usedEncoding;
	}
	
	public Boolean hasBom() {
		return hasBom;
	}
	
	public class UnicodeFileInfo {
		
		public boolean bom = false;
		public String encoding = null;
	}
	
	/**
	 * Data-structure class that can be used to write files back in the same format way they where read.
	 * I.e. with or without a UTF BOM, using a particular encoding.
	 * Use in combination with method writeBom.
	 */
	public UnicodeFileInfo getFileInfo() {
		
		UnicodeFileInfo fi = new UnicodeFileInfo();
		fi.bom = hasBom();
		fi.encoding = getEncoding();
		return fi;
	}
	
	public static byte getBomByteSize(String encoding) {
		
		if (encoding.equals("UTF-8") || encoding.equals(UTF8))
			return 3;
		if (encoding.equals("UTF-16BE") || encoding.equals("UTF-16LE")
				|| encoding.equals(UTF16) || encoding.equals(UTF16BE) || encoding.equals(UTF16LE))
			return 2;
		if (encoding.equals("UTF-32BE") || encoding.equals("UTF-32LE"))
			return 4;
		return 0;
	}

	/**
	 * Writes a UTF BOM identifier at the beginning of the file, if needed (ufi.bom is true).
	 * @param ufi
	 * The (UTF-) file configuration from when the file was opened (@see getFileInfo()).
	 */
	public static boolean writeBom(FileOutputStream bomOut, UnicodeFileInfo ufi) throws IOException {
		
		if (!ufi.bom)
			return true;
		return writeBom(bomOut, ufi.encoding);
	}

	/**
	 * Writes the UTF BOM according to the UTF encoding.
	 */
	public static boolean writeBom(FileOutputStream bomOut, String encoding) throws IOException {
		
		byte bomSize = getBomByteSize(encoding);
		if (bomSize == 0)
			return false;
		if (encoding.equals("UTF-8") || encoding.equals(UTF8)) {
			bomOut.write((int)0xEF);
			bomOut.write((int)0xBB);
			bomOut.write((int)0xBF);
			return true;
		}
		if (encoding.equals("UTF-16BE") || encoding.equals(UTF16BE)) {
			bomOut.write((int)0xFE);
			bomOut.write((int)0xFF);
			return true;
		} 
		if (encoding.equals("UTF-16LE") || encoding.equals(UTF16LE)) {
			bomOut.write((int)0xFF);
			bomOut.write((int)0xFE);
			return true;
		}
		if (encoding.equals("UTF-32BE")) {
			bomOut.write((int)0x00);
			bomOut.write((int)0x00);
			bomOut.write((int)0xFE);
			bomOut.write((int)0xFF);
			return true;
		} 
		if (encoding.equals("UTF-32LE")) {
			bomOut.write((int)0xFF);
			bomOut.write((int)0xFE);
			bomOut.write((int)0x00);
			bomOut.write((int)0x00);
			return true;
		} 
		return false;
	}

	/**
	 * Determine if a file has a UTF BOM.
	 */
	public static boolean hasBom(String fileName) {
			
			boolean fhasBom = false;
			UnicodeReader in = null;
	
			try {
				in = new UnicodeReader(new FileInputStream(fileName), defaultCharset);
				fhasBom = in.hasBom();
			} catch(IOException unused) {
	//			unused.printStackTrace();
			} finally {
				try {
					if (in != null) in.close();
				} catch (IOException unused) {}
			}
			return fhasBom;
		}

	/**
	 * Determines if a file has a certain UTF encoding.
	 * @return The UTF encoding used in the file, or the default encoding when no UTF encoding was detected. 
	 */
	public static String getEncoding(String fileName) {
			
		String encoding = defaultCharset;
		UnicodeReader in = null;
		
		try {
			in = new UnicodeReader(new FileInputStream(fileName), defaultCharset);
			encoding = in.getEncoding();
		} catch(IOException unused) {
			//			unused.printStackTrace();
		} finally {
			try {
				if (in != null) in.close();
			} catch (IOException unused) {}
		}
		return encoding;
	}
}