package net.java.dev.fwutil;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Timer;
import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.URI;
import java.util.Date;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import javax.swing.filechooser.FileSystemView;


/**
 * A collection of convenience methods and variables.
 * Most methods have a relation to the run-environment (e.g. getArgs() and
 * getOSVariable()) or files (e.g. isUnicodeFile() and addFilesToZip()).
 * <br> A number of methods supports bigStringToParameters() which
 * converts a text file with key=value lines to a hashtable.
 * Labels relies on this method to retrieve label-keys and text from
 * the label-file. 
 * @author Fred
 *
 */
public final class Util {
	
	private static Util thisInstance = new Util();
	
	/** The directory from which the application was started. */
	public static final String startDir = System.getProperty("user.dir");
	/** The name of the operating system according to Java. */
	public static final String osName = System.getProperty("os.name").toLowerCase();
	/** The file separator as used by the OS (on Windows it is \, everybody else uses /. */
	public static final char fileSeparator = System.getProperty("file.separator").toCharArray()[0];
	/** Same as fileSeparator ? */
	public static final String pathSeparator = System.getProperty("path.separator");
	/** The next line character(s) as used by the OS. */
	public static final String newline = System.getProperty("line.separator");
	/** Identifier for UTF-8 unicode. */
	public static final String utf8Encoding = "UTF-8";
	/** The home-directory of the user (e.g. home/aname). 
	 * For the user-data directory, use getDataDir(). 
	 */
	public static final String userHome = System.getProperty("user.home");
	/** A temp-directory that can be used to store data for one session. */
	public static final String tempDir = System.getProperty("java.io.tmpdir");
	/** The "My Documents" directory. 
	 * For the user-data directory, use getDataDir(). 
	 */
	public static final String userFileDir = FileSystemView.getFileSystemView().getDefaultDirectory().getAbsolutePath();
	/** The directory with configuration information for the application. */ 
	public static final String confDir = startDir + fileSeparator + "conf";
	/** The directory with images for the application. */ 
	public static final String imgDir = startDir + fileSeparator + "img";
	/** A date format that can be used in filenames. */
	public static final SimpleDateFormat simpleDateTime = new SimpleDateFormat("yyyy.MM.dd.HH.mm.ss");
	/** A short date-time format. */
	public static final SimpleDateFormat shortDateTime = new SimpleDateFormat("MM/dd HH:mm");
	/** A short date-time format for debug-mode (shows thousands of a second). */
	public static final SimpleDateFormat debugDateTime = new SimpleDateFormat("MM/dd HH:mm.ss.SSS");
	/** The short time format as preferred by the user (according to the OS). */
	public static final DateFormat userShortTime = DateFormat.getTimeInstance(DateFormat.SHORT);
	/** The short date format as preferred by the user (according to the OS). */
	public static final DateFormat userShortDate = DateFormat.getDateInstance(DateFormat.SHORT);
	/** The short date-time format as preferred by the user (according to the OS). */
	public static final DateFormat userShortDateTime = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT);
	
	/** An empty String (length 0). */
	public static final String emptyString = "";
	/** The space char as String. */
	public static final String space = " ";
	/** Amount of milli-seconds in one day. */
	public static final long oneDay = 1000 * 60 * 60 * 24;
	/** A next line character. */
	public static final char carriageReturn = net.java.dev.fwutil.StringLine.carriageReturn;
	/** Another next line characters. */
	public static final char lineFeed = net.java.dev.fwutil.StringLine.lineFeed;
	/** A marker used to separate various file-extensions in one String. */ 
	public static final String valueSeparator = ";";
	/** A marker to indicate the start of a comment/remark line. */
	public static final String commentLineStart = "//";
	/** Another marker to indicate the start of a comment/remark line. */
	public static final String commentLineStart2 = "#";
	/** A marker to indicate the continuation of a value on the next line. */
	public static final String continueTextValue = " \\";
	/** A marker to separate a key from a value in a String. */
	public static final char keyValueSeparator = '=';
	
	// Used by selectFiles()
	private static ArrayList<String> fileList;
	// Used by readFileToString. Buffersize of 128k works ok with files up to 10MB.
	private static int bufferSize = 131072;
	// Used by readFileToString. 
	private static char[] charBuffer = new char[bufferSize];
	// Timer instance.
	private static Timer timer = null;
		
	private Util() {
		super();
	}
	
	/**
	 * An application should use only one Timer to keep the amount
	 * of threads low. Here you can get it.
	 * @return One timer.
	 */
	public static Timer getTimer() {
		if (timer == null)
			timer = new Timer();
		return timer;
	}
	
	/**
	 * Recursive method to calculate the faculty of a number.
	 */
	public static long faculty(int v) {
		
		return (v == 1 ? 1 : (v * faculty(v - 1)));
	}
	
	/** Returns true if the application is running on a Windows OS. */
	public static boolean isWindowsOS() {
		
		if (osName.indexOf("winnt")>-1) return true;
		if (osName.indexOf("windows")>-1) return true;
		return false;
	}
	
	/** Returns the user specific data directory (e.g. to store user related settings
	 * of the application in.
	 */
	public static String getUserDataDir() {
		
		if ((isWindowsOS()) && (osName.indexOf("windows 9")<0)) // Windows but not Win95-98
			return getOSVariableValue("APPDATA");
		// The default:
		return userHome;
	}
	
	/** Appends dirName to the user data directory and creates the 
	 * directory so you can store application data in this directory.
	 */
	public static String getDataDir(String dirName) {
		
		/*
		 * Defaults to temp-dir when dirname is an empty string
		 */
		
		File f;
		String dataDir;
		
		if (dirName == null || dirName.equals(Util.emptyString)) {
			dataDir = Util.tempDir;
		} else {
			if (dirName.charAt(0) != fileSeparator)
				dirName = fileSeparator + dirName;
			dataDir = Util.getUserDataDir() + dirName;
		}
		while (dataDir.length()>0 && dataDir.endsWith(File.separator)) {
			dataDir = dataDir.substring(0, dataDir.length()-1);
		}
		f = new File(dataDir);
		if (!f.exists()) {
			try {
				if (!f.mkdirs()) 
					return null;
				dataDir = f.getCanonicalPath();
			} catch (IOException e) {
				System.out.println("Failed to open data directory " + dataDir);
				System.out.println(e);
				return null;
			}
		}
		return dataDir;
	}
	
	/** Returns true when the directory exists. */
	public static boolean isExistingDir(String dir) {
		
		if (dir == null || dir.equals(emptyString))
			return false;
		File f = new File(dir);
		if (f.exists() && f.isDirectory())
			return true;
		return false;
	}
	
	public static InputStream getFileStream(String fileName) {

		InputStream in = null;
		if (isValidFile(fileName)) {
			try {
				in = new FileInputStream(new File(fileName));
			} catch (FileNotFoundException e) {
				in = null;
			}
		}
		if (in == null)
			in = ClassLoader.getSystemResourceAsStream(fileName);
		if (in == null)
			in = Thread.currentThread().getContextClassLoader().getResourceAsStream(fileName);
		if (in == null)
			in = thisInstance.getClass().getClassLoader().getResourceAsStream(fileName);
		return in;
	}
	
	/** Utility function that copies bytes from in to out. */
	public static void copyStreams(InputStream in, OutputStream out) throws IOException {
		
		byte[] buf = new byte[8192];
		int len;
		while((len = in.read(buf)) > 0) {
			out.write(buf, 0, len);
		}
	}

	/** Returns true if the file exists. Returns false if the
	 * given filename is a directory.
	 */
	public static boolean isValidFile(File f) {
		
		if (f == null)
			return false;
		if (!f.exists())
			return false;
		if (f.isFile())
			return true;
		return false;
	}

	
	/** Returns true if the file exists. Returns false if the
	 * given filename is a directory.
	 */
	public static boolean isValidFile(String fileName) {
		
		if (fileName == null || fileName.equals(emptyString))
			return false;
		return (isValidFile(new File(fileName)));
	}
	
	/** Returns true if the file exists. Returns false if the
	 * given filename is a directory.
	 */
	public static boolean isValidFile(URI uri) {
		
		if (uri == null)
			return false;
		return isValidFile(new File(uri));
	}

	/** Returns true if the file is a Unicode text-file. */
	public static String isUnicodeFile(String fileName) {
	
		if (!isValidFile(fileName))
			return null;
		int byteValue = 1;
		BufferedInputStream in = null;
		String encoding = null;
		try {
			in = new BufferedInputStream(new FileInputStream(fileName));
			byteValue = in.read();
			if (byteValue == (int)0xEF) {
				if (in.read() == (int)0xBB)
					if (in.read() == (int)0xBF)
						encoding = "UTF-8";
			} else if (byteValue == (int)0xFE) {
				if (in.read() == 0xFF) 
					encoding = "UTF-16BE";
			} else if (byteValue == (int)0xFF) {
//				System.out.println("Here2 " + byteValue + " " + ((int)0xFF));					
				if (in.read() == (int)0xFE) {
					byteValue = in.read();
//					System.out.println("Here3" + byteValue);					
					if ((byteValue == (int)0x00) && (in.read() == (int)0x00))
						encoding = "UTF-32LE";
					else 
						encoding = "UTF-16LE";
				}
			} else if (byteValue == (int)0x00) {
				if (in.read() == (int)0x00)
					if (in.read() == (int)0xFE)
						if (in.read() == (int)0xFF)
							encoding = "UTF-32BE";
			}
		} catch (IOException unused) {
		} finally {
			try {
				if (in != null) in.close();
			} catch (IOException unused) {}
		}
//		System.out.println(encoding);
		return encoding;
	}
	
	/** Returns true if the file is not a text file. */
	public static boolean isBinaryFile(String fileName) {
		/*
		 * Determines if a file is a binary file by looking for a 0-byte
		 * in the first 2048 bytes of the file (characters are always non-zero bytes).
		 * This is not a perfect method but should work for most cases.
		 * Note that false is also returned in case of error.
		 */
		
		if (isUnicodeFile(fileName) != null)
			return false;
		int byteValue = 1;
		BufferedInputStream in = null;
		try {
			in = new BufferedInputStream(new FileInputStream(fileName));
			int count = 1;
			while (byteValue != 0 && byteValue != -1 && count < 2048) {
				byteValue = in.read();
				count++;
			}
		} catch (IOException unused) {
		} finally {
			try {
				if (in != null) in.close();
			} catch (IOException unused) {}
		}
		if (byteValue == 0)
			return true;
		return false;
	}
	
	/** Deleted a file. */
	public static boolean deleteFile(String fileName) {
		
		if (!isValidFile(fileName))
			return false;
		File f = new File(fileName);
		return f.delete();
	}

	/** Executes the command in the OS. */
	public static Process Execute(String command) {
		
		Process p = null;
		Runtime r = Runtime.getRuntime();
		
		if (isWindowsOS()) { 
			try {
				if (osName.indexOf("windows 9")>-1) {  // Win95-98
					p = r.exec("command /c " + command);
				} else { // Winnt, 2000, XP
					p = r.exec("cmd /c " + command);
				}
			} catch (IOException ioe) {
				;
			}
		} else { // assume Unix
			try {
				p = r.exec("/bin/bash -c " + command);
			} catch (IOException ioe) {
				;
			}
		}
		return p;
	}
	
	/** Closes all open streams and destroys the proces. */ 
	public static void closeProcess(Process p) {
		
		if (p == null) return;
		if (p.getErrorStream() != null)
			try { p.getErrorStream().close(); } catch (IOException ioe) {;}
		if (p.getInputStream() != null)
			try { p.getInputStream().close(); } catch (IOException ioe) {;}
			try { p.getOutputStream().close(); } catch (IOException ioe) {;}
		p.destroy();
	}
	
	/** Gets the response from the OS after executing the
	 * command in the OS.
	 */
	public static String getOSResponse(String command) {
		
		String response = null;
		Process p;
		
		p = Execute(command);
		if (p != null) {
			try {
				BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));
				p.waitFor();
				response = in.readLine();
			} catch (Throwable t) {
			}
			closeProcess(p);
		}
		return response;
	}
	
	/** Gets the OS value for the given OSVariable (e.g. Path). */
	public static String getOSVariableValue(String OSVariable) {
		
		if (OSVariable == null || OSVariable.equals(emptyString)) return null;
		String value = null;
		if (isWindowsOS()) {
			value = getOSResponse("echo %" + OSVariable + "%");
		} else { // Assume unix
			value = getOSResponse("$" + OSVariable);
		}
		return value;
	}
	
	/** Waits on System.out for a response from the user. 
	 * Returns the input from the user.
	 */
	public static String getResponse() {
	
		String response = null;
		System.out.flush();
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		try {
			response = in.readLine();
		} catch (IOException unused) {}
		return response;
	}
	
	/**
	 * Converts the args from the command line to a hashtable
	 * with keys and values. 
	 * Requires program arguments in the following format:
	 * <br> -key="A value"
	 * <br> or
	 * <br> -key
	 * <br> A parameter that does not start with the - sign is stored as key.
	 * <br> Example:
	 * <br> -debug -fileName="C:\Program Files\somefile.txt" AnotherKey
	 * @param args The arguments from main.
	 * @return A hastable with keys and (possibly) values.
	 */
	public static Hashtable<String, String> getArgs(String[] args) {
		
		Hashtable<String, String> params = new Hashtable<String, String>();
		
		String key;
		String value;
		for (int i = 0; i < args.length; i++) {
			key = "";
			value = "";
			if (args[i].startsWith("-") && args[i].indexOf("=")<0) {
				key = args[i].substring(1, args[i].length());
			}
			if (args[i].startsWith("-") && args[i].indexOf("=")>-1) {
				key = args[i].substring(1, args[i].indexOf("="));
				value = args[i].substring(args[i].indexOf("=")+1, args[i].length());
			}
			if (!args[i].startsWith("-")) {
				key = args[i];
			}
			params.put(key, value);
		}
		return params;		
	}

	/** Returns current date-time in simpleDateTime format. */
    public static String simpleDateTime() {
		
		return simpleDateTime(new Date(System.currentTimeMillis()));
	}
	
	/** Returns the given date in simpleDateTime format. */
	public static String simpleDateTime(Date d) {
		
		return simpleDateTime.format(d);
	}

	/** Returns current date-time in shortDateTime format. */
	public static String shortDateTime() {
		
		return shortDateTime(new Date(System.currentTimeMillis()));
	}

	/** Returns the given date in shortDateTime format. */
	public static String shortDateTime(Date d) {
		
		return shortDateTime.format(d);
	}
	
	/** Returns current date-time in debugDateTime format. */
	public static String debugDateTime() {
		
		return debugDateTime(new Date(System.currentTimeMillis()));
	}
	
	/** Returns the given date in debugDateTime format. */
	public static String debugDateTime(Date d) {
		
		return debugDateTime.format(d);
	}
	
	/** Returns the given date in userShortTime format. */
	public static String userShortTime(Date d) {
		
		return userShortTime.format(d);
	}

	/** Returns the given date in userShortDate format. */
	public static String userShortDate(Date d) {
		
		return userShortDate.format(d);
	}

	/** Returns the given date in userShortDateTime format. */
	public static String userShortDateTime(Date d) {
		
		return userShortDateTime.format(d);
	}

/*	
 * 
 * Replaced by BrowserLauncher
 * 
 * 
	public static boolean openURL(String url) {
		
		if (url == null || url.length() == 0) {
			return false;
		}
		String browser = webBrowserDefault;
		if (webBrowser.equals(emptyString)) {
			if (isWindowsOS()) {
				// Use default browser on Windows, fails on Unix
				browser = "rundll32 url.dll,FileProtocolHandler";
			}
		} else
			browser = webBrowser;
		String[] execs = new String[] {browser, url};
		try {
			Runtime.getRuntime().exec(execs);
		} catch (IOException unused) {
			return false;
		}
		return true;
	}
*/

	/**
	 * Returns an arraylist of filename given the selection criteria.
	 * @param fromDir The directory to select files in.
	 * @param fileNameEndsWith The extension(s) the files must have 
	 *  (if null, all files are selected). Extensions are separator by 
	 *  Util.valueSeparator (e.g. ".exe;.com").
	 *  @param readOnly If FALSE, only selects writable files. Set to true
	 *  to ignore the read-only property of files.
	 *  @param maxSize The maximum size of files to select. If this
	 *  is smaller then 0, file size is ignored during selection. 
	 */
	public static ArrayList<String> selectFiles(String fromDir,  
			String fileNameEndsWith, boolean readOnly, long maxSize) {
		
		ArrayList<String> fileDirList = new ArrayList<String>();
		String[] files = new File(fromDir).list();
		ArrayList<String> fExtensions = new ArrayList<String>();
		if (fileNameEndsWith != null && !fileNameEndsWith.equals(emptyString)) {
			StringTokenizer fExt = new StringTokenizer(fileNameEndsWith, valueSeparator, false);
			while (fExt.hasMoreTokens())
				fExtensions.add(fExt.nextToken());
		}
		File f;
		boolean fileOK;
		boolean extensionFound;
		for (int i = 0; i < files.length; i++) {
			f = new File(fromDir + fileSeparator + files[i]);
			if (!f.isDirectory()) {
				fileOK = true;
				if (fExtensions.size() > 0) {
					extensionFound = false;
					int j = 0;
					while (j < fExtensions.size() && !extensionFound) {
						extensionFound = files[i].endsWith(fExtensions.get(j));
						j++;
					}
					fileOK = extensionFound;
				}
				if (fileOK && !readOnly && !f.canWrite())
					fileOK = false;
				if (maxSize >= 0)
					fileOK = fileOK && (f.length() < maxSize);
				if (fileOK)
					fileDirList.add(f.getAbsolutePath());
			}
		}
		if (fileDirList.size() == 0)
			return null;
		return fileDirList;
	}
	
	/** Returns all directories in the given directory. */
	public static ArrayList<String> getDirs(String fromDir) { 

		ArrayList<String> dirList = new ArrayList<String>();
		String[] files = new File(fromDir).list();
		File f;
		
		for(int i = 0; i < files.length; i++) {
			f = new File(fromDir + fileSeparator + files[i]);
			if (f.isDirectory())
				dirList.add(f.getAbsolutePath());
		}
		if (dirList.size() == 0)
			return null;
		return dirList;
	}

	/**
	 * Used by selectFiles if recursive is true.
	 */
	private static void recursiveSelectFiles(String fromDir, boolean includeDirs,
			String fileNameEndsWith, boolean readOnly, long maxSize) {
		
		ArrayList<String> files = null;
		ArrayList<String> dirs = null;

		dirs = getDirs(fromDir);
		if (dirs != null) {
			for (int i = 0; i < dirs.size(); i++) {
				recursiveSelectFiles(dirs.get(i), includeDirs, fileNameEndsWith,
						readOnly, maxSize);
			}
		}
		files = selectFiles(fromDir, fileNameEndsWith, readOnly, maxSize);
		if (dirs != null && includeDirs)
			fileList.addAll(dirs);
		if (files != null)
			fileList.addAll(files);
	}
	
	/** Selects files from fromDir.
	 * @param recursive If true, all sub-directories are included.
	 * @param includeDirs If true, includes directories in the returned list.
	 * @param fromDir The directory to select files in.
	 * @param fileNameEndsWith The extension(s) the files must have 
	 *  (if null, all files are selected). Extensions are separator by 
	 *  Util.valueSeparator (e.g. ".exe;.com").
	 * @param readOnly If FALSE, only selects writable files. Set to true
	 *  to ignore the read-only property of files.
	 * @param maxSize The maximum size of files to select. If this
	 *  is smaller then 0, file size is ignored during selection.
	 *  @return A list if files (and directories) identified by
	 *  their absolute path. 
	 */
	public static ArrayList<String> selectFiles(String fromDir, boolean recursive, boolean includeDirs, 
			String fileNameEndsWith, boolean readOnly, long maxSize) {
		
		fileList = new ArrayList<String>();
		
		if (!isExistingDir(fromDir))
			return null;
		if (!recursive) {
			ArrayList<String> files = null;
			ArrayList<String> dirs = null;
			files = selectFiles(fromDir, fileNameEndsWith, readOnly, maxSize);
			if (includeDirs)
				dirs = getDirs(fromDir);
			if (files != null)
				fileList.addAll(files);
			if (dirs != null)
				fileList.addAll(dirs);
			if (fileList.size() == 0)
				return null;
			return fileList;
		}
		recursiveSelectFiles(fromDir, includeDirs, fileNameEndsWith,
				readOnly, maxSize);
		if (fileList.size() == 0)
			return null;
		return fileList;
	}
	
	/** Reads a text file into a String using the default encoding. */
	public static String readFileToString(String fileName) {

		return readFileToString(fileName, UnicodeReader.defaultCharset);
	}
	
	public static String readFileToString(InputStream fin, String encoding) {
		
		if (fin == null)
			return null;
		StringBuffer out = new StringBuffer();
		boolean error = false;
		UnicodeReader in = null;
		
		try {
			in = new UnicodeReader(fin, encoding);
			int charsRead;
			while ((charsRead = in.read(charBuffer, 0, bufferSize)) != -1)
				out.append(charBuffer, 0, charsRead);
//System.out.println("enc: " + in.getEncoding() + " bom: " + in.hasBom());

		} catch(IOException unused) {
//unused.printStackTrace();
				error = true;
		} finally {
			try {
				if (in != null) in.close();
			} catch (IOException unused) {
				error = true;
			}
		}
		if (!error)
			return out.toString();
		else
			return null;
	}
		
	/** Reads a text file into a String using the given encoding. 
	 * See UnicodeReader for Unicode encoding names
	 */
	public static String readFileToString(URI fileURI, String encoding) {
	
		if (!isValidFile(fileURI))
			return null;
		InputStream fin = null;
		try {
			fin = fileURI.toURL().openStream();
		} catch (Exception e) {}
		return readFileToString(fin, encoding);
	}

	/** Reads a text file into a String using the given encoding. 
	 * See UnicodeReader for Unicode encoding names
	 */
	public static String readFileToString(String fileName, String encoding) {
	
		if (!isValidFile(fileName))
			return null;
		File f = new File(fileName);
		FileInputStream fin = null;
		try {
			fin = new FileInputStream(f);
		} catch (FileNotFoundException e) {}
		return readFileToString(fin, encoding);
	}

	/** Checks if the line starts with a comment/remark marker. */
	public static boolean isCommentLine(String line) {
		
		return (line.startsWith(commentLineStart) || line.startsWith(commentLineStart2) ? true : false);
	}
	
	/** Returns the key and value from the text line.
	 * If trimValues is true, spaces surrounding the value are removed.
	 * @param line The text-line
	 * @param keyValue The key and value from the text-line
	 * @param trimValues If true, spaces surrounding the value are removed.
	 * @return True if a key and value was found in the text-line.
	 * @throws Exception "Not a key-value line".
	 */
	public static boolean getKeyValue(String line, String[] keyValue, boolean trimValues) throws Exception {
		
		String key = null;
		String value = null;
		if (!isCommentLine(line)) {
			int valueIndex = line.indexOf(keyValueSeparator);
			if (valueIndex != -1) {
				key = line.substring(0, valueIndex);
				key = key.trim();
				value = null;
				if (valueIndex + 1 < line.length())
					value = line.substring(valueIndex + 1, line.length());
				if (value != null) {
					if (trimValues)
						value = value.trim();
					keyValue[0] = key;
					keyValue[1] = value;
					return true;
				} else {
					value = emptyString;
					keyValue[0] = key;
					keyValue[1] = value;
					return true;
				}
			} else {
				// Unknown line
				throw new Exception("Not a KeyValue line: " + newline + line);
			}
		} // line != commentLine
		return false;
	}
	
	/** Continues to read a value that consists of multiple lines. */
	public static String continueReadMultiLineValue(String value, StringLine.FileStringIterator sfi, final int bigStringLength) {
		
		int multiLineCount = 1;
		String line = value; 
		while (line.endsWith(continueTextValue) && sfi.startFrom < bigStringLength) {
			String nextLine = StringLine.getNextLine(sfi);
			if (!isCommentLine(nextLine)) {
				line = line.substring(0, line.length() - continueTextValue.length()) + Util.newline + nextLine;
				multiLineCount++;
			}
		}
		if (Settings.appDebug)
			Log.log(Integer.toString(multiLineCount) + " lines read for multi line value.");
		return line;
	}
	
	/** 
	 * Converts text to key-value pairs.
	 * @param bigString The big String containing the text.
	 * @param trimValues If true, spaces surrounding the value are removed.
	 * @return The key-value pairs from the String.
	 */
	public static Hashtable<String, String> bigStringToParameters(final String bigString, boolean trimValues) {
		
		if (bigString == null || bigString.length() == 0)
			return null;
		int bigStringLength = bigString.length();
		Hashtable<String, String> parameters = new Hashtable<String, String>();
		StringLine.FileStringIterator sfi = StringLine.getNewFSI(bigString, 0);
		String line = null;
		while (sfi.startFrom < bigStringLength) {
			line = StringLine.getNextLine(sfi);
			if (line != null) {
				String[] keyValue = new String[2];
				try {
					if (getKeyValue(line, keyValue, trimValues)) {
						if (keyValue[1].endsWith(continueTextValue)) {
							if (Settings.appDebug)
								Log.log("Continue reading multi line value for " + keyValue[0] + " ...");
							keyValue[1] = continueReadMultiLineValue(keyValue[1], sfi, bigStringLength);
						}
						parameters.put(keyValue[0], keyValue[1]);
					}
				} catch (Exception e) {
					if (Settings.appDebug) {
						Log.logWarning("Line number " + sfi.lineCount + " - " + e);
					}
				}
			} else {
				if (sfi.startFrom < bigString.length()) {
					String s = "Line number "  + sfi.lineCount + " - " +
						"Unexpected null-line encountered";
					Log.logWarning(s);
				}
			} // line != null
		} // while not all lines processed
		if (Settings.appDebug) {
			Log.log(sfi.lineCount + " lines read containing " + parameters.size() + " keys.");
		}
		return parameters;
	}
	
	/**
	 * Writes a big string to a text file.
	 * @param out The text to write to file.
	 * @param fileName The name of the file.
	 * @param encoding The encoding to use, see UnicodeReader.
	 * @param writeBom If true, will write a BOM-marker at the beginning of the file if applicable.
	 * @return True if no error occurred.
	 */
	public static boolean writeStringToFile(String out, String fileName, String encoding, boolean writeBom) {

		boolean error = false;
		OutputStreamWriter fileOut = null;
		
		try {
			char[] charOut = out.toCharArray();
			FileOutputStream bomOut = new FileOutputStream(fileName);
			if (writeBom)
				UnicodeReader.writeBom(bomOut, encoding);
			fileOut = new OutputStreamWriter(bomOut, encoding);
			fileOut.write(charOut, 0, charOut.length);
		} catch(IOException unused) {
				error = true;
		} finally {
			try {
				if (fileOut != null) fileOut.close();
			} catch (IOException unused) {
				error = true;
			}
		}
		return !error;
	}
	
	/**
	 * Writes a big string to a text file using the default encoding.
	 * @param out The text to write to file.
	 * @param fileName The name of the file.
	 * @return True if no error occurred.
	 */

	public static boolean writeStringToFile(String out, String fileName) {
		
		return writeStringToFile(out, fileName, UnicodeReader.defaultCharset, false);
	}
	
	/** Writes a serializable object to file using the XMLEncoder. */
	public static boolean writeObject(String fileName, Object o) {

		boolean error = false;
		FileOutputStream out = null;
		XMLEncoder s = null;
		
		try {
			out = new FileOutputStream(fileName);
			s = new XMLEncoder(out);
			s.writeObject(o);
			s.flush();
		} catch (IOException ieo) {
//			ieo.printStackTrace();
			error = true;
		} finally {
			try {
				if (s != null) s.close();
				if (out != null) out.close();
			} catch (IOException ieo) {
				error = true;
			}
		}			
		return (!error);
	}
	
	
	/** Reads a serializable object from file using the XMLDecoder. */
	public static Object readObject(String fileName) {
		
		if (!isValidFile(fileName))
			return false;
		
		Object o = null;
		FileInputStream in = null;
		XMLDecoder s = null;
		boolean error = false;
		
		try {
			in = new FileInputStream(fileName);
			s = new XMLDecoder(in);
			o = s.readObject();
		} catch (IOException ieo) {
//			ieo.printStackTrace();
			error = true;
		} finally {
			try {
				if (s != null) s.close();
				if (in != null) in.close();
			} catch (IOException ieo) {
				error = true;
			}
		}
		if (error) return null;
		return o;
	}
	
	/**
	 * Stores files in a zip-file. Uses Log to write messages. 
	 * <br> NOTE: an existing zip-file is DELETED before the operation starts.
	 * @param rootDir The directory to read files from.
	 * @param fnames The names of the files to store in the zip-file. If a path is 
	 * included in the file-name the path must be relative to rootDir.
	 * @param zipFile The absolute path and name of the zip-file.
	 * @return False if an error occurred.
	 */
	public static boolean addFilesToZip(String rootDir, String[] fnames, String zipFile) {
		
		Util.deleteFile(zipFile);
		ZipOutputStream zout = null;
		try {
			zout = new ZipOutputStream(new FileOutputStream(zipFile));
		} catch (FileNotFoundException e) {
			Log.logWarning("Could not create zip-file: " + zipFile + Util.newline + Log.exception2String(e));
			return false;
		}
		if (Settings.appDebug)
			Log.log("Opening new zip file: " + zipFile);
		if (!rootDir.endsWith(String.valueOf(fileSeparator)))
			rootDir = rootDir + fileSeparator;
		boolean fileAdded = false;
		InputStream in = null;
		try {
			String fname;
			byte bbuffer[] = new byte[bufferSize];
			for (int i = 0; i < fnames.length; i++) {
				fname = fnames[i];
				if (fname.startsWith(String.valueOf(fileSeparator)))
					fname = fname.substring(1, fname.length());
				if (!isValidFile(rootDir + fname)) {
					Log.logWarning("Cannot add file to zipfile: " + rootDir + fname);
				} else {
					ZipEntry ze = new ZipEntry(fname.replace(File.separatorChar,'/'));
					in = new FileInputStream(rootDir + fname);
					zout.putNextEntry(ze);
					int len = 0;
					while ( (len = in.read(bbuffer)) != -1) {
						zout.write(bbuffer, 0, len);
					}
					zout.closeEntry();
					fileAdded = true;
					in.close();
					if (Settings.appDebug)
						Log.log("Added file to zipfile: " + rootDir + fname);
				}				
			}
		} catch (IOException ioe) {
			Log.logWarning("Exception during zip-operation: " + Log.exception2String(ioe));
		} finally {
			try {
				if (in != null) in.close();
				if (zout != null) zout.close();
			} catch (IOException ioe2) { 
				Log.logWarning("Exception during closing of zip-operation: " + Log.exception2String(ioe2));
			}
		}
		if (Settings.appDebug) 
			Log.log("Zipfile closed.");
		return fileAdded;
	}

	/**
	 * String.replace() only replaces 1 occurrence of the the string
	 * that needs to be replaced. This procedure continues with
	 * replacement until no occurrence is found anymore.
	 * String.replaceAll uses regular expression that I do not know. 
	 * @param toRemove The text to remove
	 * @param inText The text to remove text in.
	 * @return The inText without the toRemove caracters.
	 */
	public static String removeAll(final String toRemove, final String inText) {
		
		String olds = inText;
		String sclean = olds.replace(toRemove, Util.emptyString);
		while (!sclean.equals(olds) && sclean.length() > 0) {
			olds = sclean;
			sclean = olds.replace(toRemove, Util.emptyString);
		}
		return sclean;
	}
}
