package net.java.dev.fwutil;

import java.io.Serializable;
import java.util.HashMap;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Point;
import java.awt.Component;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.Toolkit;

import net.java.dev.hvlayout.HVProps;


/**
 * An XMLEncoder-persistable class that stores or sets window dimensions.
 * If applyWindowDimension is used before a Window is packed (pack())and shown 
 * <br> and setWindowDimension is used before a Window is closed,
 * <br> then the Window's size and position will be remembered and restored
 * (Settings is used to store and retrieve the values).
 * <br> This geatly enhances user experience: the user will not need to
 * resize and reposition windows to get a good view of it's contents.
 * Once set, it stays set.
 * <p>
 * If applyWindowDimension is used for the first time,
 * the window is positioned at the middle of the screen or 
 * relative to it's owner. Size should be determined by a call to pack().
 * <p>
 * Methods specific to Java beans contract are marked with JB.
 * @author Fred
 *
 */
public class WindowDimension implements Serializable {

	private static final long serialVersionUID = 1L;
	/**
	 * JB - values that must be stored must be private.
	 */
	private int x;
	private int y;
	private int width;
	private int height;
	private boolean maximized;
	private static HashMap<String, WindowDimension> wdData;
	public static final Dimension screenSizePrimary = Toolkit.getDefaultToolkit().getScreenSize();
	public static final Dimension screenSizeTotal = new Dimension();
	static {
		Rectangle virtualBounds = new Rectangle();
		GraphicsEnvironment ge = GraphicsEnvironment.
		getLocalGraphicsEnvironment();
		GraphicsDevice[] gs =
			ge.getScreenDevices();
		for (int j = 0; j < gs.length; j++) { 
			GraphicsDevice gd = gs[j];
			GraphicsConfiguration[] gc =
				gd.getConfigurations();
			for (int i=0; i < gc.length; i++) {
				virtualBounds =
					virtualBounds.union(gc[i].getBounds());
			}
		}
		screenSizeTotal.width = virtualBounds.width;
		screenSizeTotal.height = virtualBounds.height;
	}

	public static final String wdTable = "WindowsDimensionData";
	
	/** Used as default width when this class is instantiated.. */
	public static final int defaultWindowWidth = 500;
	/** Used as default height when this class is instantiated.. */
	public static final int defaultWindowHeight = 350;

	/**
	 * JB - a constructor with no parameters.
	 *
	 */
	public WindowDimension() {

		this(100, 100, defaultWindowWidth, defaultWindowHeight, false);
	}
	
	/**
	 * JB (I think) - a constructor with all parameters.
	 * @param x Horizontal start position
	 * @param y Vertical start position
	 * @param width Width of window
	 * @param height Height of window
	 * @param maximized Window is maximized or not
	 */
	public WindowDimension(int x, int y, int width, int height, boolean maximized) {
		
		super();
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
		this.maximized = maximized;
	}

	/** JB - getter for private int x */
	public int getX() {
		
		return x;
	}
	
	/** JB - setter for x */
	public void setX(int x) {
		
		this.x = x;
	}

	/** JB - getter for y */
	public int getY() {
		
		return y;
	}
	
	/** JB - setter for y */
	public void setY(int y) {
		
		this.y = y;
	}

	/** JB - getter for Width, note that the name of the getter must
	 * be in line with the name of the priavte variable (private int width --> getWidth)
	 */
	public int getWidth() {
		
		return width;
	}

	/** JB - setter for width */
	public void setWidth(int width) {
		
		this.width = width;
	}

	/** JB - getter for height*/
	public int getHeight() {
		
		return height;
	}

	/** JB - setter for height*/
	public void setHeight(int height) {
		
		this.height = height;
	}
	
	/** JB - getter for maximized*/
	public boolean getMaximized() {
		
		return maximized;
	}
	
	/** JB - setter for maximized*/
	public void setMaximized(boolean maximized) {
		
		this.maximized = maximized;
	}
	
	/** Returns the width and height as dimension. */
	public Dimension getDimension() {
		
		return new Dimension(width, height);
	}
	
	/** Returns the x and y location as Point. */
	public Point getLocation() {
		
		return new Point(x, y);
	}
	
	/** Positions the window at the middle of the screen. */
	public static void positionAtMiddleOfScreen(Window c) {
		
		c.setLocation((int)((screenSizePrimary.getWidth() - c.getSize().getWidth()) / 2),
					  (int)((screenSizePrimary.getHeight() - c.getSize().getHeight()) / 2));
	}
	
	/** Gets the window dimensions and stores it in an instance of this class. */
	private static void component2WindowDimension(Window c, 
			WindowDimension wd) {
		
		boolean storeDimensions = true;
		if (c instanceof Frame) {
			if ( (((Frame)c).getExtendedState() &  Frame.MAXIMIZED_BOTH) 
						== Frame.MAXIMIZED_BOTH) {
			wd.setMaximized(true); 
			storeDimensions = false;
			// Cannot set values from maximized window
			// this will cause problems (e.g. window not showing).
			} else if ((((Frame)c).getExtendedState() &  Frame.NORMAL) 
					!= Frame.NORMAL) {
				storeDimensions = false;
				// Iconified windows have no size, so do not store it.
				// It is not likely that when the window is called again,
				// the user wants it to show iconified.
			}
		} 
		if (storeDimensions) {
			wd.setMaximized(false);
			wd.setX(new Double(c.getLocation().getX()).intValue());
			wd.setY(new Double(c.getLocation().getY()).intValue());
			wd.setWidth(c.getWidth());
			wd.setHeight(c.getHeight());
		}
	}
	
	/** Applies stored window dimensions to the window if they exist.
	 * If window dimensions do not exist, they are fetched and stored in memory
	 * and the window is placed at the middle of the screen.
	 */
	public static void applyWindowDimension(String windowName, Window c) {
		applyWindowDimension(windowName, c, null);
	}
	
	//@SuppressWarnings("unchecked")
	/** Applies stored window dimensions to the window if they exist.
	 * If window dimensions do not exist, they are fetched and stored in memory.
	 * If an owner exists, the window location is set relative to the owner.
	 */
	@SuppressWarnings("unchecked")
	public static void applyWindowDimension(String windowName, Window c, Component owner) {
		
		wdData = (HashMap<String, WindowDimension>)Settings.get(wdTable); 
		if (wdData == null) {
			wdData = new HashMap<String, WindowDimension>();
			Settings.put(wdTable, wdData);
		}
		WindowDimension cwd = wdData.get(windowName);
		if (cwd == null) {
			cwd = new WindowDimension();
			if (owner == null)
				positionAtMiddleOfScreen(c);
			else {
				c.setLocationRelativeTo(owner);
//				c.setLocation(owner.getLocationOnScreen());
			}
			component2WindowDimension(c, cwd);
			wdData = (HashMap<String, WindowDimension>)Settings.get(wdTable); 
			if (wdData == null) {
				wdData = new HashMap<String, WindowDimension>();
				Settings.put(wdTable, wdData);
			}
			wdData.put(windowName, cwd);
		} else {
			if (cwd.getMaximized()) {
				if (c instanceof Frame)
					((Frame)c).setExtendedState(Frame.MAXIMIZED_BOTH);
			} else {
				if (owner == null) {
					if (cwd.getX() < screenSizeTotal.width - cwd.getWidth() && cwd.getY() < screenSizeTotal.height- cwd.getHeight()) {
						c.setBounds(cwd.getX(), cwd.getY(), cwd.getWidth(), cwd.getHeight());
					} else {
						c.setSize(cwd.getWidth(), cwd.getHeight());
						positionAtMiddleOfScreen(c);
					}
				} else {
					if (owner.getX()  + HVProps.defaultLineHeight < screenSizeTotal.width - cwd.getWidth() 
							&& owner.getY() + HVProps.defaultLineHeight < screenSizeTotal.height- cwd.getHeight()) {
						c.setBounds(owner.getX() + HVProps.defaultLineHeight, 
								owner.getY() + HVProps.defaultLineHeight,
								cwd.getWidth(), cwd.getHeight());
					} else {
						c.setSize(cwd.getWidth(), cwd.getHeight());
						positionAtMiddleOfScreen(c);
					}
				}
			}
//			c.setLocation(cwd.getLocation());
//			c.setSize(cwd.getDimension());
		}
//		c.pack();
//		c.repaint();
		
		// Sometimes Windows are not repainted when they are showing and are resized 
		// by this method (a call to repaint() does not work). 
		// A crude but effective work-around is to
		// use setVisible to update Window sizes.
		// But only for a frame, a modal dialog will hang on setVisible()
		if (c.isVisible() && !(c instanceof Dialog)) {
			c.setVisible(false);
			c.setVisible(true);
		}
	}
	
	//@SuppressWarnings("unchecked")
	/**
	 * Stores the window dimensions in an instance of this class.
	 */
	@SuppressWarnings("unchecked")
	public static void setWindowDimension(String windowName, Window c) {
		
		wdData = (HashMap<String, WindowDimension>)Settings.get(wdTable); 
		if (wdData == null) {
			wdData = new HashMap<String, WindowDimension>();
			Settings.put(wdTable, wdData);
		}
		WindowDimension cwd = wdData.get(windowName);
		if (cwd == null)
			cwd = new WindowDimension();
		component2WindowDimension(c, cwd);
		wdData.put(windowName, cwd);
	}
}

