/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.hvlayout;


import javax.swing.JComponent;
//import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.UIManager;
import javax.swing.border.EmptyBorder;
import javax.swing.plaf.FontUIResource;

import java.awt.BorderLayout;
import java.awt.Canvas;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Toolkit;

import java.awt.Insets;
import java.awt.GridBagConstraints;
import java.util.Enumeration;

/**
 * This singleton class contains (default) values and methods that should be used when
 * working with {@code HLayout} and {@code VLayout}. See for an example {@code FormUI}.
 * The functions cover the following:
 * <br>  - Determining and settings sizes for components (class {@code FontDependentSizes}
 * and method {@code setSize()}. Note that all functions in this class will
 * attempt to use the component's font size to determine the size of the component.
 * But if this cannot be done, the static values defaultFontSize, defaultLineHeight,
 * defaultButtonWidth and defaultLineWidth are used.
 * <br>  - Methods that manipulate component grow and shrink behavior 
 * methods {@code (doNot)Shrink} and {@code (doNot)Grow}.
 * <br>  - Methods that manipulate the position of components
 * ({@code getCanvasLine()} and {@code getEmptyLine()}.
 * <br>  - Methods that add scrollers around components
 * ({@code getScroller()} and {@code setDefaultScrollerSize()}.
 * <br>  - Methods that manipulate the font that is shown on screen
 * ({@code get/setDefault(UI)Font(Size)}).
 *  
 * @author Fred
 */
public final class HVProps {

	/** Inset with 0 space all around. */
	public static final Insets noInsets = new Insets(0, 0, 0, 0);

	/** Default insets, used in VBox(insets) and FormUI. */
	public static final Insets defaultInsets = new Insets(2, 2, 2, 2);

	/** A border taking no space at all. */
	public static final EmptyBorder noBorder = new EmptyBorder(noInsets);

	/** A default border with default insets. Used by VBox(null). */
	public static final EmptyBorder defaultBorder = new EmptyBorder(defaultInsets);

	/**
	 * In some cases the default line height can be made smaller.
	 * This is true for labels that are placed directly beneath each other and
	 * it is true for rows in JTable.
	 * <br> On the other hand, it might point at an unknown layout issue.
	 */
	public static final int smallHeightCorrection = 4;
	/**
	 * Screensize here for convenience and to determine the maximum width in setSize()
	 * and for a good constant in {@code VLayout.maximumLayoutSize} which
	 * also uses the maximum width. 
	 */
	public static final Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
	
	/**
	 * The font size. Value is aligned with the font size as used by the UIManager
	 * (via method {@code getDefaultUIFontSize()}) upon initialization of 
	 * this singeleton class.
	 */
	private static int defaultFontSize = 12;

	/**
	 * The line height: the default height for buttons, labels, textfields, etc.
	 * Scales with font size.
	 * It so happens that 12 results in 20 (see also {@code setFontDependentSizes()}.
	 */
	public static int defaultLineHeight = 20;

	/**
	 * The width of buttons, scales with font size.  
	 */
	public static int defaultButtonWidth = defaultLineHeight * 5;

	/**
	 * The preferred width of textfields, scales with font size. 
	 */
	public static int defaultLineWidth = defaultButtonWidth * 2;
	
	/**
	 * The default horizontal gap between components as used by HLayout
	 */
	public static int defaultHGap = 2;

	/**
	 * The default vertical gap between components as used by VLayout
	 */
	public static int defaultVGap = 2;

	// Boolean to determine if VLayout has been initialized.
	private static boolean fontInit = false;

	private static HVProps thisInstance = new HVProps();

	private HVProps() {
		super();
		if (!fontInit) {
			// Synchronize with UIManager fonts
			final FontDependentSizes fsizes = new FontDependentSizes(
					getDefaultUIFontSize());
			defaultFontSize = getDefaultUIFontSize();
			defaultLineHeight = fsizes.lineHeight;
			defaultButtonWidth = fsizes.buttonWidth;
			defaultLineWidth = fsizes.lineWidth;
			fontInit = true;
		}
	}
	
	/**
	 * Needed to get an instance of FontDependentSizes
	 * via e.g. HVProps.getInstance().new FontDependentSizes(null);
	 * @return The only instance of HVProps
	 */
	public static HVProps getInstance() {
		return thisInstance;
	}
	
	/**
	 * Calculates button width, line height and line width
	 * given the (component's) font size.
	 */
	public class FontDependentSizes {

		public transient int buttonWidth;
		public transient int lineHeight;
		public transient int lineWidth;

		public FontDependentSizes(Component c) {

			super();
			if (c.getFont() == null)
				setFontDependentSizes(defaultFontSize);
			else
				setFontDependentSizes(c.getFont().getSize());
		}

		public FontDependentSizes(int fontSize) {

			super();
			if (fontSize == 0)
				setFontDependentSizes(defaultFontSize);
			else
				setFontDependentSizes(fontSize);
		}

		private void setFontDependentSizes(final int fontSize) {

			// tweakHeight is required for font sizes 11, 8 etc., see formula
			int tweakHeight = 0; 
			if (fontSize % 3 != 0)
				tweakHeight = 1;
			// Formula looks weird but works for small and large font sizes
			// The part (defaultFontHeight / 3) adds height of g and h etc.
			// i.e. extra space above and below the character
			lineHeight = fontSize + fontSize / 3 +
			4 + tweakHeight;
			// The 4 is not related to defaultInsets or something
			// it is a constant. smallHeightCorrention could be related.
			buttonWidth = lineHeight * 4;
			lineWidth = buttonWidth * 2;
		}
	}

	/** All components from {@code FormUI} call this function to ensure proper 
	behaviour for HVLayout.
	<br> The component height is fixed. 
	(maxHeight will be set to preferred Height).
	@param c		The component that should be sized according to the other parameters
	@param width	The width of the component, set to buttonWidth if 0
	@param height	The height of the component, set to lineHeight if 0
	@param maxLength	If true, component will not grow wider then the specified width
	 */
	public static void setSize(final Component c, int width, int height, 
			final boolean maxLength) {

		final FontDependentSizes fsizes = thisInstance.new FontDependentSizes(c);
		if (width == 0)
			width = fsizes.buttonWidth;
		if (height == 0)
			height = fsizes.lineHeight;
		c.setPreferredSize(new Dimension(width, height));
		if (width >= fsizes.buttonWidth)
			c.setMinimumSize(new Dimension(fsizes.buttonWidth, height));
		else 
			c.setMinimumSize(new Dimension(width, height));
		if (maxLength)
			c.setMaximumSize(new Dimension(width, height));
		else
			c.setMaximumSize(new Dimension(screenSize.width, height));
	}

	/** Ensures a component cannot shrink horizontally. */
	public static void doNotShrinkWidth(final Component c) {
		c.setMinimumSize(new Dimension(c.getPreferredSize().width, c.getMinimumSize().height));
	}
	/** Ensures a component can shrink to half it's preferred size horizontally. */
	public static void doShrinkWidth(final Component c) {
		c.setMinimumSize(new Dimension(c.getPreferredSize().width / 2, c.getMinimumSize().height));
	}
	/** Ensures a component cannot shrink vertically. */
	public static void doNotShrinkHeigth(final Component c) {
		c.setMinimumSize(new Dimension(c.getMinimumSize().width, c.getPreferredSize().height));
	}
	/** Ensures a component can shrink vertically to half it's preferred size. */
	public static void doShrinkHeigth(final Component c) {
		c.setMinimumSize(new Dimension(c.getMinimumSize().width, c.getPreferredSize().height / 2));
	}
	/** Ensures a component cannot shrink. */
	public static void doNotShrink(final Component c) {
		c.setMinimumSize(new Dimension(c.getPreferredSize().width, c.getPreferredSize().height));
	}
	/** Ensures all components in the container cannot shrink. */
	public static void doNotShrinkContainer(final Container c) {
		for(int i = 0; i < c.getComponentCount(); i++)
			doNotShrink(c.getComponent(i));
	}
	/** Ensures a component cannot grow horizontally. */
	public static void doNotGrowWidth(final Component c) {
		c.setMaximumSize(new Dimension(c.getPreferredSize().width, c.getMaximumSize().height));
	}
	/** Ensures a component can grow horizontally to twice it's preferred size. */
	public static void doGrowWidth(final Component c) {
		c.setMaximumSize(new Dimension(c.getPreferredSize().width * 2, c.getMaximumSize().height));
	}
	/** Ensures a component cannot grow vertically. */
	public static void doNotGrowHeigth(final Component c) {
		c.setMaximumSize(new Dimension(c.getMaximumSize().width, c.getPreferredSize().height));
	}
	/** Ensures a component can grow vertically to twice it's preferred size. */
	public static void doGrowHeigth(final Component c) {
		c.setMaximumSize(new Dimension(c.getMaximumSize().width, c.getPreferredSize().height * 2));
	}
	/** Ensures a component cannot grow. */
	public static void doNotGrow(final Component c) {
		c.setMaximumSize(new Dimension(c.getPreferredSize().width, c.getPreferredSize().height));
	}
	/** Ensures all components in the container cannot grow. */
	public static void doNotGrowContainer(final Container c) {
		for(int i = 0; i < c.getComponentCount(); i++)
			doNotGrow(c.getComponent(i));
	}

	/**
	 * Returns an empty line, can be used for spacing vertically.
	 */
	public static Canvas getCanvasLine(final boolean maxLength) {

		final Canvas cv = new Canvas();
		setSize(cv, 0, 0, maxLength);
		return cv;
	}

	/**
	 * Returns an empty line that mimics the given HLayout container.
	 * Can be used to keep alignment of fields in place (virtual divider). 
	 * See {@code AddressBookDemo} for an example.
	 */
	public static Container getEmptyLine(final Container box) {

		final Container lb = new Container();
		int hgap = HVProps.defaultHGap; 
		if (box.getLayout() instanceof HLayout)
			hgap = ((HLayout)box.getLayout()).hGap;
		lb.setLayout(new HLayout(hgap));
		Component c;
		Canvas cv;
		for (int i = 0; i < box.getComponentCount(); i++) {
			c = box.getComponent(i);
			cv = new Canvas();
			cv.setMinimumSize(new Dimension(c.getMinimumSize().width, c.getMinimumSize().height));
			cv.setPreferredSize(new Dimension(c.getPreferredSize().width, c.getPreferredSize().height));
			cv.setMaximumSize(new Dimension(c.getMaximumSize().width, c.getMaximumSize().height));
			lb.add(cv);
		}
		return lb;
	}
	
	/**
	 * Adds a scroller around the view-component. 
	 * Uses the view-component's preferred width and height to
	 * to set the scroller's sizes. 
	 * @param view The component that is shown using a scroller if needed
	 * @param setMin When true sets the minimum width equal to the preferred width so that
		the scroller will not shrink beyond preferred width.
	 * @return The scroller showing the view.
	 */
	public static JScrollPane getScroller(final JComponent view, 
			final boolean setMin) {

		final JScrollPane scroller = new JScrollPane(view);
		setDefaultScrollerSize(scroller, setMin);
		return scroller;
	}

	/**
	 * Adds a scroller around the view-component, can be used with H/VBox
	 * Uses the given width and height (in lines) to set the preferred sizes of the scroller.
	 * @param view The component that is shown using a scroller if needed
	 * @param prefWidth The preferred width of the scroller (value 0 is defaulted to lineWidth)
	 * @param lines The height of the scroller in number of lines (value 0 is defaulted to 2)
	 * @param setMin sets the minimum width equal to the preferred width so that
		the scroller will not shrink beyond preferred width. Usefull for JLists for example.
	 * @return The scroller
	 */
	public static JScrollPane getScroller(final JComponent view, int prefWidth,
			final int lines, final boolean setMin) {

		final JScrollPane scroller = new JScrollPane(view);
		if (prefWidth == 0)
			prefWidth = defaultLineWidth;
		setDefaultScrollerSize(scroller, prefWidth,
				lines * defaultLineHeight, setMin);
		return scroller;
	}

	/** Sets the setDefaultScrollerSize using the
	 * preferred width and height of the scoller's viewport.
	 * Applicable to, for example, JList.
	 * See further setDefaultScrollerSize(scroller, prefWidth, prefheight, setMin)
 	 * @param setMin sets the minimum width equal to the preferred width so that
	 * the scroller will not shrink beyond preferred width. Usefull for JLists for example.
	 */
	public static void setDefaultScrollerSize(final JScrollPane scroller, 
			final boolean setMin) {

		final Dimension dpref = scroller.getViewport().getView().getPreferredSize();
		setDefaultScrollerSize(scroller, dpref.width, dpref.height, setMin);
	}

	/**
	 * Returns a scroller that only shows scroll-bars when they are needed.
	 * @param scroller The scroll-pane to set the sizes for.
	 * @param prefWidth The preferred width of the scroll-pane. Value 0 is defaulted to buttonWidth.
	 * @param prefHeight The preferred height of the scroll-pane. Value 0 is defaulted to 2 * lineHeight.
	 * @param setMin If true, scroller will not shrink vertically. Usefull for JLists for example.
	 */
	public static void setDefaultScrollerSize(final JScrollPane scroller, int prefWidth,
			int prefHeight, final boolean setMin) {

		final Component scrollable = scroller.getViewport().getView();
		final FontDependentSizes fsizes = thisInstance.new FontDependentSizes(scrollable);
		/*
		 * Text areas should not display a tooltip 
		 * when the text area is inside a scroller.
		 */
		if (prefWidth == 0)
			prefWidth = fsizes.buttonWidth;
		if (prefHeight == 0)
			prefHeight = 2 * fsizes.lineHeight;
		final int minHeight = Math.min(2 * fsizes.lineHeight, prefHeight);
		int minWidth = Math.min(fsizes.buttonWidth, prefWidth);
		if (setMin)
			minWidth = prefWidth;
		// Following statement is only required for layouts that do not take the Insets into account
		// scroller.getViewport().setViewSize(scroller.getViewport().getView().getPreferredSize());
		final Insets scrollerInsets = scroller.getInsets();
		final int scrollerInsetsWidth = scrollerInsets.left + scrollerInsets.right;
		final int scrollerInsetsHeight = scrollerInsets.top + scrollerInsets.bottom;
		// The tweak with the insets is required because they are
		// also used in the preferred size. Omitting the inserts results
		// in mostly useless scrollbars appearing when the size shrinks. 
		scroller.setMinimumSize(new Dimension(minWidth + scrollerInsetsWidth, 
				minHeight + scrollerInsetsHeight));
		// The tweak with the insets is essential here, else you will always
		// see a scrollbar even when the contents of the scroller 
		// reach preferred or maximum size
		scroller.setPreferredSize(new Dimension(prefWidth + scrollerInsetsWidth, 
				prefHeight + scrollerInsetsHeight));
	}
	
	/**
	 * If a scroller always has a horizontal scrollbar, then use
	 * this function to adjust preferred size of the scroller
	 * so that the contents of the scroller are shown in full
	 * when preferred size is used in the layout.
	 * @param scroller The scroller with the always showing horizontal scrollbar.
	 */
	public static void addHScrollbarHeight(final JScrollPane scroller) {
		
		final int sbheight = scroller.getHorizontalScrollBar().getMaximumSize().height;
		scroller.setPreferredSize(new Dimension(scroller.getPreferredSize().width, 
				scroller.getPreferredSize().height + sbheight));
	}

	/**
	 * If a scroller always has a vertical scrollbar, then use
	 * this function to adjust preferred size of the scroller
	 * so that the contents of the scroller are shown in full
	 * when preferred size is used in the layout.
	 * @param scroller The scroller with the always showing vertical scrollbar.
	 */
	public static void addVScrollbarWidth(final JScrollPane scroller) {
		
		final int sbwidth = scroller.getVerticalScrollBar().getMaximumSize().width;
		scroller.setPreferredSize(new Dimension(scroller.getPreferredSize().width
				+ sbwidth, scroller.getPreferredSize().height ));
	}

	/** The defaultFontSize as used by this layout manager.*/
	public static int getDefaultFontSize() {
		return defaultFontSize;
	}

	/** Aligns the defaultFontSize and other 
	 * static dependent values in this class with the given fontSize.
	 */
	public static void setDefaultFontSize(final int fontSize) {

		final FontDependentSizes fsizes = thisInstance.new FontDependentSizes(fontSize);
		defaultFontSize = fontSize;
		defaultLineHeight = fsizes.lineHeight;
		defaultButtonWidth = fsizes.buttonWidth;
		defaultLineWidth = fsizes.lineWidth;
	}
	
	/** Returns The default font as used by Swing's UIManager, Label.font
	 * Label.font is most likely always available and set
	 * by setDefaultUIFont().
	 */
	public static Font getDefaultUIFont() {
		return (Font)UIManager.get("Label.font");
		// and NOT: return (Font)UIManager.getLookAndFeelDefaults().get("Label.font");
	}

	/** The default font size as used by Swing's UIManager,
	 * see also getDefaultUIFont()*/
	public static int getDefaultUIFontSize() {
		return getDefaultUIFont().getSize();
	}
	
	/** Returns An instance of a mono-spaced font (like Courier) with the
	 * default font size.
	 */
	public static Font getDefaultUIFontMonoSpaced() {
		return new Font("Monospaced", 0, HVProps.getDefaultFontSize());
	}

	/** Sets all used fonts for displaying to the specified font df
	 * Aligns values in HVProps. 
	 * <br>Use with care! 
	 * */
	@SuppressWarnings("unchecked")
	public static void setDefaultUIFont(final Font df) {

		setDefaultFontSize(df.getSize());
		final Enumeration keys = UIManager.getLookAndFeelDefaults().keys();
		while (keys.hasMoreElements()) {
			final Object key = keys.nextElement();
			final Object value = UIManager.get(key);
			if (value instanceof Font) {
				final Font ifont = (Font)value;
				final Font ofont = new FontUIResource(df.getName(), ifont.getStyle(), df.getSize());
				UIManager.put(key, ofont);
			}
		} 
	}

	/** Convenience method for setDefaultUIFont(Font) */
	public static void setDefaultUIFont(final String name, final int style, 
			final int size) {
		final Font f = new Font(name, style, size);
		setDefaultUIFont(f);
	}

	/** Adjust all fonts for displaying to the given size.
	 * Aligns values in HVProps. 
	 * See also TestUI (shows window with normal or big font using this method) 
	 * <br>Use with care!
	 */ 
	@SuppressWarnings("unchecked")
	public static void setDefaultUIFontSize(final int size) {

		setDefaultFontSize(size);
		final Enumeration keys = UIManager.getLookAndFeelDefaults().keys();
		while (keys.hasMoreElements()) {
			final Object key = keys.nextElement();
			final Object value = UIManager.get(key);
			if (value instanceof Font) {
				final Font ifont = (Font)value;
				final Font ofont = ifont.deriveFont((float)size);
				UIManager.put(key, ofont);
			}
		}    
	}
	
	/** Returns a string with all fonts used by Swing's UIManager.
	 * See the bottom text-box in {@code DemoUI} that shows the string
	 * returned by this method.
	 */
	@SuppressWarnings("unchecked")
	public static String getUIFonts() {
		
		final StringBuffer fonts = new StringBuffer(128);
		fonts.append("Default font: ");
		fonts.append(HVProps.getDefaultUIFont().toString());
		final Enumeration keys = UIManager.getLookAndFeelDefaults().keys();
		while (keys.hasMoreElements()) {
			final Object key = keys.nextElement();
			final Object value = UIManager.get(key);
			if (value instanceof Font) {
				final Font ifont = (Font)value;
				fonts.append(System.getProperty("line.separator") + key.toString() + " " + ifont.getName() + " " + ifont.getStyle() + " " + ifont.getSize());				
			}
		}
		return fonts.toString();
	}
	
	/** Convenience method, uses gaps as set by this layoutmanager.*/
	public static BorderLayout getBorderLayout() {
		return new BorderLayout(defaultHGap, defaultVGap);
	}
	/** Convenience method, uses gaps as set by this layoutmanager.*/
	public static GridLayout getGridLayout(final int rows, final int cols) {
		return new GridLayout(rows, cols, defaultHGap, defaultVGap);
	}
	/** Convenience method, uses gaps as set by this layoutmanager.*/
	public static FlowLayout getFlowLayout(final int alignment) {
		return new FlowLayout(alignment, defaultHGap, defaultVGap);
	}
	/** Convenience method, uses defaultInsets as set by this layoutmanager.*/
	public static GridBagConstraints getGridBagC(final int gridx, final int gridy, 
			final int gridwidth, final int gridheight, final int anchor, 
			final int fill) {
		return new GridBagConstraints(gridx, gridy, gridwidth, gridheight,
				0.0, 0.0, anchor, fill, HVProps.defaultInsets, 0, 0);
	}
}
