/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.hvlayout;

import java.awt.ComponentOrientation;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;
import javax.swing.border.EmptyBorder;

/**
 * A window that demonstrates various capabilities of HVLayout.
 * This screen can be used to test modifications of HVLayout but
 * also to show the effects of the different settings. E.g.:
 * <br> - HVTextArea is used in two places with different parameters.
 * <br> - Tooltips should only appear when text is truncated.
 * <br> - Buttons and text lines are aligned to the left, right and center.
 * <br> - etc.
 * <br> Some of the capabilities shown when the top three
 * checkboxes are activated:
 * <br> - Internationalization support: click "Switch orientation" to
 * see how components are layed out in right-to-left reading countries.
 * <br> - Font size adaptability: click "Switch big font" to see
 * how component' sizes change to accomodate a bigger font.
 * <br> - Scrolling behavior: click "Add scroller" to see how boxes with
 * scrollers behave.
 * @author Fred
 *
 */
public class TestUI {

	public static JFrame frame;
	// Determines to use a big font size for display 
	public static boolean showWithBigFont = false;
	// Boolean Left to right, determining line axis orientation
	public static boolean ltor = true;
	public static boolean showWindowScroller = false;
	public static boolean showDemoScroller = false;
	public static int standardFontSize;
	
	// This is a singleton class.
	private TestUI() {
		super();
	}
	
	public static ActionListener swicthOrientation = new ActionListener() {
		public void actionPerformed(final ActionEvent e){
			ltor ^= true; // Faster then ltor = !ltor
			showDemoForm();
		}
	};
	
	public static ActionListener bigFont = new ActionListener() {
		public void actionPerformed(final ActionEvent e){
			showWithBigFont ^= true;
			showDemoForm();
		}
	};

	public static ActionListener showScrollerAL = new ActionListener() {
		public void actionPerformed(final ActionEvent e){
			showWindowScroller ^= true;
			showDemoForm();
		}
	};

	public static ActionListener showDemoScrollerAL = new ActionListener() {
		public void actionPerformed(final ActionEvent e){
			showDemoScroller ^= true;
			showDemoForm();
		}
	};

	public static void showDemoForm() {
		
		JScrollPane scroller;
		frame.setVisible(false);
		// Clear the frame from any previous filling 
		frame.getContentPane().removeAll();
		// Top level container, a VBox which aligns from top to bottom
		VBox mainPane = new VBox(HVProps.defaultInsets);
		mainPane.setBorder(new EmptyBorder(HVProps.defaultInsets));
		// Set default font size, necessary when showWithBigFont is turned off
		HVProps.setDefaultUIFontSize(standardFontSize);
		if (showWithBigFont)
			HVProps.setDefaultUIFontSize((int)(standardFontSize * 1.5));
		// Following four variables are re-used
		HBox lb, lb2;
		VBox left;
		VBox middle;
		// Checkbox that when clicked, rebuilds screen with reverse orientation
		JCheckBox cb = FormUI.getCheckBox(swicthOrientation, "Switch orientation", 0);
		cb.setSelected(!ltor);
		mainPane.add(cb);
		cb = FormUI.getCheckBox(bigFont, "Switch big font", 0);
		cb.setSelected(showWithBigFont);
		mainPane.add(cb);
		cb = FormUI.getCheckBox(showScrollerAL, "Add scroller to window", 0);
		cb.setSelected(showWindowScroller);
		mainPane.add(cb);
		lb = new HBox(HBox.CENTER);
		final JTextArea centeredText = FormUI.getTextArea("Centered multi label example\n " +
				"Second line\n Third and last line of multi label", false);
		centeredText.setFont(centeredText.getFont().deriveFont(Font.BOLD));
		lb.add(centeredText);
		mainPane.add(lb);
		
		mainPane.add(FormUI.getTextArea("Below are three form boxes in one line box\n" +
				"The left part containing the labels\n" +
				"is outlined opposite line axis and has fixed size\n" +
				"The most right part is showing some small buttons", false));
		cb = FormUI.getCheckBox(showDemoScrollerAL, "Add scroller", 0);
		cb.setSelected(showDemoScroller);
		mainPane.add(cb);
		
		left = new VBox();
		// Following labels are outlined opposite the line axis 
		// Note that thanks to VBox (and also HBox) the labels
		// are also outlined opposite the line axis when orientation is switched
		// from left to right and vice versa
		left.add(FormUI.getLabel("Name", HVProps.defaultButtonWidth , true, true));
		left.add(FormUI.getLabel("Street", HVProps.defaultButtonWidth, true, true));
		left.add(FormUI.getLabel("City", HVProps.defaultButtonWidth, true, true));
		lb = new HBox();
		lb.add(left);
		middle = new VBox();
		middle.add(FormUI.getTextField("Frederik Wiers", 0));
		middle.add(FormUI.getTextField("Marco Polostraat", 0));
		middle.add(FormUI.getTextField("Amsterdam", 0));
		lb.add(middle);
		VBox right = new VBox();
		lb2 = new HBox();
		lb2.add(FormUI.getSmallButton("1"));
		lb2.add(FormUI.getTextField("One text field", 0));
		lb2.add(FormUI.getSmallButton("11"));
		right.add(lb2);
		lb2 = new HBox();
		lb2.add(FormUI.getSmallButton("2"));
		lb2.add(FormUI.getTextField("Another text field", 0));
		lb2.add(FormUI.getSmallButton("21"));
		right.add(lb2);
		lb2 = new HBox();
		lb2.add(FormUI.getSmallButton("3"));
		lb2.add(FormUI.getTextField("Last text field", 0));
		lb2.add(FormUI.getSmallButton("31"));
		right.add(lb2);
		lb.add(right);
		if (showDemoScroller) {
			scroller = HVProps.getScroller(lb, false);
			mainPane.add(scroller);
		} else
			mainPane.add(lb);

		mainPane.add(FormUI.getLabel("Below a text area with default size of 3 lines", 0));
		JTextArea text = FormUI.getTextArea("", true);
		text.setText("This is just some demo text to see what happens with the text area when resizing is active.\n" +
				"This is the second line.\n" +
				"The text area has a preferred size of 3 lines.\n" +
				"And a default width.\n" +
				"This is the 5th line, more then preferred.\n" +
				"But still visible if resized.\n");
		mainPane.add(HVProps.getScroller(text, 0, 3, false));
		mainPane.add(FormUI.getTextArea("Lines that are centered or reversed (opposite lines axis)\n" +
				"keep their preferred width (else you might not see the difference with normal outlined lines)", false));
		lb = new HBox(HBox.CENTER);
		lb.add(FormUI.getButton("Demo button"));
		lb.add(FormUI.getTextField("Demo text here centered in screen", 0));
		lb.add(FormUI.getSmallButton("End"));
		mainPane.add(lb);
		mainPane.add(FormUI.getLabel("Below the same line but not centered", 0));
		lb = new HBox();
		lb.add(FormUI.getButton("Demo2 button with a long description"), new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				throw new NullPointerException("Do exception done");
			}
		});
		lb.add(FormUI.getTextField("Demo2 text here centered in screen", 0));
		lb.add(FormUI.getSmallButton("End2"));
		mainPane.add(lb);
		mainPane.add(FormUI.getLabel("Below three line-boxes with different orientations containing buttons", 0));
		lb2 = new HBox();
		lb = new HBox();
		lb.add(FormUI.getButton("Normal outline"));
		lb2.add(lb);
		lb = new HBox(HBox.CENTER);
		lb.add(FormUI.getButton("Center outline"));
		lb2.add(lb);
		//mainPane.add(UIComp.getLabel("Following buttons are outlined opposite the line axis (reversed)", 0));
		lb = new HBox(HBox.OPPOSITELINEAXIS);
		lb.add(FormUI.getButton("Opposite 1"));
		lb.add(FormUI.getButton("Opposite 2"));
		//lb.add(UIComp.getButton("outlined"));
		lb2.add(lb);
		mainPane.add(lb2);
		
		mainPane.add(FormUI.getTextArea("Below a text area with default size of 6 lines\n" +
				"note that the text area below will be twice the size of the text area above", false));
		text = FormUI.getTextArea(HVProps.getUIFonts(), false);
		mainPane.add(HVProps.getScroller(text, 0, 6, false));
		if (showWindowScroller) {
			scroller = HVProps.getScroller(mainPane, false);
			//scroller.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
			mainPane = new VBox();
			mainPane.add(scroller);
		}

		frame.setContentPane(mainPane);
		if (!ltor)
			frame.applyComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
		frame.pack();
		if (frame.getSize().height > HVProps.screenSize.height - 50)
			frame.setSize(frame.getSize().width, HVProps.screenSize.height - 50);
		frame.setVisible(true);	
	}

	public static void main(String[] args) {

		// Call to UIManager.setLookAndFeel must be done before a call to HVProps.
		// If UIManager is used after HVProps is used, synchronize HVProps with
		// the UIManager by calling HVProps.setDefaultFontSize(FormLayout.getDefaultUIFontSize());
		try {
            UIManager.setLookAndFeel(
                UIManager.getSystemLookAndFeelClassName());
        } catch (Exception unused) { }
        // standardFontSize must be fetched after the above call to UIManager
        // because that installs the operating system default font.
        // The first instance of HVProps will align 
        // all static variables in HVProps with the default font
        // from UIManager.
        standardFontSize = HVProps.getDefaultUIFontSize();
        frame = new JFrame("HVLayout demo form");
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        ltor = frame.getComponentOrientation().isLeftToRight();
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				showDemoForm();
            }
        });
	}
}
