/*
 * This software is licensed under the GNU Lesser GPL, see
 * http://www.gnu.org/licenses/ for details.
 * You can use this free software in commercial and non-commercial products,
 * as long as you give credit to this software when you use it.
 * This softwware is provided "as is" and the use of this software is 
 * completely at your own risk, there are absolutely no warranties.
 * 
 */

package net.java.dev.hvlayout;


import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager2;
import java.util.ArrayList;

/**
 * A layout-manager that aligns components vertically and shrinks or
 * grows components vertically as specified by their minimum, preferred and maximum sizes.
 * Use {@code VBox.add(Component)} to put each added component 
 * beneath the previous one and let this layout manager do the vertical resizing.
 * <p>
 * The component's sizes must be set properly for this layout manager 
 * to work. Use {@code HVProps.setSize()} to accomplish this.
 * Use {@code HVProps} many default values to change to appearance of the screen. 
 * <p> 
 * Components shrink vertically to their minimum size when their preferred size 
 * is larger then their minimum size.<br>
 * Likewise, components will grow vertically when their maximum size is greater
 * then their preferred size.<br>
 * Also, components shrink and grow relative to how much other components want to
 * shrink or grow. For example, if one text area (A) needs to be twice as big as 
 * another text area (B), then simply setting the preferred height of B to 
 * twice the preferred height of A will do the trick. See for example 
 * {@code TestUI} which calls 
 * <br> {@code HVProps.getScroller(text, 0, 3, false)} and 
 * <br> (@code HVProps.getScroller(text, 0, 6, false)}
 * to show this effect. 
 * <p>
 * This class determines the size for components vertically. The {@code HLayout} 
 * class determines the size for components horizontally and
 * can be used together with HBox to order components in
 * the reading direction, from the middle or opposite the reading direction.
 * <p>
 * Other layour managers should have no trouble with this layout manager. 
 * In other words, {@code VBox and HBox} can be part of a component that uses  
 * BorderLayout or Flowlayout, etc.
 * <p>
 * Add instances of {@code VBox} (which uses {@code VLayout}) and {@code HBox} (which uses
 * {@code HLayout}) to the content pane of frames and dialogs (and/or make
 * a {@code VBox} with insets the content pane).
 * <p>
 * Developers note:
 * VLayout and HLayout must be kept in sync to work.  
 */
public class VLayout implements LayoutManager2 { 
	
	// The components to align vertically
	private transient ArrayList<Component> lines; 
	// MaximumSize of this form
	private transient Dimension dmax = null;
	// MinimumSize of this form
	private transient Dimension dmin = null;
	// PreferredSize of this form
	private transient Dimension dpref = null;
	// Used in calculation to shrink component, set by minimumLayoutSize
	private transient int varMinHeight = 0;
	// Height from lines that do not grow in height (have height.pererred >= height.max)
	// This value may only be used after a call to preferredLayoutSize
	private transient double staticMaxHeight = 0;
	
	/** Size of vertical gap between two components in a form (vertically outlined) */ 
	public transient int vGap;

	/**
	 * Contructor to layout components from top to bottom.
	 * The gap between components defaults to HVProps.defaultVGap.
	 *
	 */
	public VLayout() {
		
		this(HVProps.defaultVGap);
	}
	
	/**
	 * Contructor to layout components from top to bottom.
	 * @param vGap The gap between the components. 
	 */
	public VLayout(int vGap) {
		
		lines = new ArrayList<Component>();
		this.vGap = vGap;
	}
	
	/** Adds a component at the end of the list or 
	 * if place is a valid index (Integer class required)
	 * adds the component at the specified
	 * index-place (0 is the first place). 
	 * <br> LineBox uses this method for a call to add(component, index)
	 * @param c The component to add
	 * @param place The place in the list of components for this layout. 
	 */
	public void addLayoutComponent(final Component c, final Object place) {

		if (place instanceof Integer) {
			final int index = ((Integer)place).intValue();
			if (index > -1 && (index < lines.size()))
				lines.add(index, c);
			else
				lines.add(c);
		} else {
			lines.add(c);
		}
	}

	/**
	 * Adds component c to the end of the list of components for this layout.
	 */
	public void addLayoutComponent(final String arg0, final Component arg1) {

		lines.add(arg1);
	}

	/** Set the general alignment to a value that corresponds to
	the line axis orientation, i.e.
	<br> left top for left to right and
	<br> right top for right to left
	*/
	public float getLayoutAlignmentX(final Container parent) {
		final boolean leftToRight = parent.getComponentOrientation().isLeftToRight();
		float f = 0.25f;
		if (!leftToRight)
			f = 0.75f; 
		return f;
	}

	/** See also getLayoutAlignmentX
	 * defaults to "near the top"
	 */
	public float getLayoutAlignmentY(final Container arg0) {
		// near the top
		return 0.25f;
	}
	
	/**
	 * The total gap height, used in resizing calculations.
	 */
	private int gapHeight() {
		
		if (!lines.isEmpty())
			return (vGap * (lines.size() - 1));
		return 0;
	}
	

	/** This method ensures that temporary values can be used.
	Only this function can give the signal that a re-calculation has to be done
	<br> According to BoxLayout this should be synchronized
	*/
	public synchronized void invalidateLayout(final Container arg0) {

		dmax = null;
		dmin = null;
		dpref = null;
//System.out.println("Invalidate lineLayout");		
	}
	
	/**
	 * Removes component c from the list of components for this layout.
	 */
	public void removeLayoutComponent(final Component arg0) {

		boolean found = false;
		int i = 0;
		while (!found && i < lines.size()) {
			found = (lines.get(i) == arg0);
			i++;
		}
		if (found)
			lines.remove(--i);
			// and NOT lines.remove(i--)
	}

	/** Return minimum size (width and height) of this container
	 * The minimum size of this layout calculated by summing up
	 * all minimum sizes of the components in this layout.
	 * <br> The container will refuse to draw smaller.
	 * <br> Also sets varMinHeight which is used in the layoutContainer method.
	 */
	public Dimension minimumLayoutSize(final Container target) {
		
		if (dmin != null)
			// By not returning dmin itself we ensure that dmin values cannot be changed from the outside
			return new Dimension(dmin.width, dmin.height);
		// maxWidth will determine the required width for this form
		int maxWidth = 0;
		// height will determine the reuired height for this form
		int height = 0;
		// varMinHeight setermines the amount of space that can be 
		// resized during shrinking from components
		varMinHeight = 0;
		// According to other layout managers, this is how you ensure that
		// access to component/line sizes is synchronized 
		synchronized (target.getTreeLock()) {
			Dimension cmin;
			Dimension cpref;
			for(int i = 0; i < lines.size(); i++) {
				cmin = lines.get(i).getMinimumSize();
				cpref = lines.get(i).getPreferredSize();
				height += cmin.height;
				maxWidth = Math.max(maxWidth, cmin.width);
				if (cpref.height > cmin.height)
					varMinHeight += (cpref.height - cmin.height);
			}
		}
		final Dimension d = new Dimension(maxWidth, height + gapHeight());
		dmin = new Dimension(d.width, d.height);
		return d;
	}

	/**
	 * The preferred size of this layout calculated by summing up
	 * all preferred sizes of the components in this layout.
	 */
	public Dimension preferredLayoutSize(final Container target) {
		if (dpref != null)
			return new Dimension(dpref.width, dpref.height);
		
		int maxWidth = 0;
		int height = 0;
		staticMaxHeight = 0;
		synchronized (target.getTreeLock()) {
			Dimension cpref;
			Dimension cmax;
			for(int i = 0; i < lines.size(); i++) {
				cpref = lines.get(i).getPreferredSize();
				cmax = lines.get(i).getMaximumSize();
				height += cpref.height;
				maxWidth = Math.max(maxWidth, cpref.width);
				if (cpref.height >= cmax.height)
					staticMaxHeight += cpref.height;
			}
		}
		final Dimension d = new Dimension(maxWidth, height + gapHeight());
		staticMaxHeight += gapHeight();
		dpref = new Dimension(d.width, d.height);
		return d;
	}

	/** Return the maximum size (width and height) of this container
	 * The maximum size of this layout calculated by summing up
	 * all maximum sizes of the components in this layout.
	 * <br> The container will refuse to draw larger
     */
	public Dimension maximumLayoutSize(final Container target) {
		if (dmax != null)
			return new Dimension(dmax.width, dmax.height);

		int maxWidth = 0;
		int height = 0;
		synchronized (target.getTreeLock()) {
			Dimension cmax;
			for(int i = 0; i < lines.size(); i++) {
				cmax = lines.get(i).getMaximumSize();
				height += cmax.height;
				maxWidth = Math.max(maxWidth, cmax.width);
			}
		}
		final Dimension d = new Dimension(maxWidth, height + gapHeight());
		dmax = new Dimension(d.width, d.height);
		return d;
	}

	/** 
	Determines size and position of all components/lines
	according to the available size given via the target parameter.
	Minimum and maximum size of components will be respected 
	so they may be non-visible if the target is too small 
	(component falls outside of Window).
	The frame.pack() funtion usually prevents this. */
	public void layoutContainer(final Container target) {

		// The target sizes to respect:
		final Insets targetInsets = target.getInsets();
		// The maximum height space to divide over the components
		final int maxHeight = target.getHeight() - targetInsets.top - targetInsets.bottom;
		// The maximum width space to divide over the components
		// targetInsets.left is taken in account when setting the size of the component (see c.setBounds below)
		int maxWidth = target.getWidth()- targetInsets.right;
		
		// Smallest determines if minimum size for all should be ysed
		boolean smallest;
		// Shrink determines if the lines/components need to shrink in height
		boolean shrink;
		// Grow determines if the lines/components need to grow in height
		boolean grow;
		synchronized (target.getTreeLock()) {
			final Dimension dminimum = minimumLayoutSize(target); // sets varMinHeight 
			final Dimension dpreferred = preferredLayoutSize(target); // sets staticMaxHeight
			// Respect minimum width
			maxWidth = Math.max(maxWidth, dminimum.width);
			// Check if minimum size is required
			smallest = maxHeight <= dminimum.height;
			// Check if shrinking is required, varMinHeight > 0 determines if component can shrink
			shrink = (!smallest) && (maxHeight < dpreferred.height) && (varMinHeight > 0);
			// Check if growing is required, dpreferred.height  > staticMaxHeight determines if components can grow 
			grow = !smallest && !shrink && (maxHeight > dpreferred.height) && (dpreferred.height  > staticMaxHeight);
			// heightFactor will be used in the calculation to make
			// a component/line shrink or grow
			double heightFactor = 0;
			if (shrink)
				// This formula should be similar to the one below
				// but that did not work. The current one does, but requires some extra calculation
				// by calling variableHeightMin.
				// Determines ratio of space to divide : maximum space to divide  
				heightFactor = (double)(maxHeight - dminimum.height) / varMinHeight;
			if (grow)
				// Determines ratio of space to divide : growth ratio  
				heightFactor = (double)(maxHeight - dpreferred.height) / (dpreferred.height - staticMaxHeight);

//System.out.println("maxH: " + maxHeight + " shrink: " + shrink + " hFactor: " + heightFactor + 
//		" varMin: " + varMinHeight);			
			// lineHeight determines where the new vertical component position is 
			double lineHeight = targetInsets.top;
			// Determnies the height of the component
			int cheight;
			Dimension cmin;
			Dimension cpref;
			int cmaxwidth;
			for (int i = 0; i < lines.size(); i++) {
				final Component c = lines.get(i);
				cmin = c.getMinimumSize();
				cpref = c.getPreferredSize();
				cmaxwidth = c.getMaximumSize().width;
				// hasStaticHeight determines if the componet can grow or shrink
				// true is default value for smallest = true
				boolean hasStaticHeight = true;
				if (!smallest) {
					if (shrink && cmin.height < cpref.height)
						hasStaticHeight = false;
					// This is the only time when getMaximumSize is called
					if (grow && cpref.height < c.getMaximumSize().height)
						hasStaticHeight = false;
				}
				// Set default value for component/line height
				cheight = (smallest ? cmin.height : cpref.height);
				if (shrink && !hasStaticHeight) {
					// Why the calculation could not be the one below, I do not know.
					// cheight = (int)(cmin.height + (cmin.height * heightFactor));
					// But this works properly with heightFactor:
					cheight = (int)(cmin.height + ((cpref.height - cmin.height) * heightFactor));
				}
				if (grow && !hasStaticHeight) {
					cheight = (int)(cpref.height + (cpref.height * heightFactor));
				}
				cmaxwidth = (cmaxwidth < maxWidth ? cmaxwidth : maxWidth);
				c.setBounds(targetInsets.left, (int)lineHeight, cmaxwidth, (int)cheight);
				lineHeight += cheight + vGap;
			}
		}
	}
}
