package nl.intercommit.mario.tools;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.rmi.RMISecurityManager;
import java.util.Enumeration;
import java.util.Properties;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.xml.DOMConfigurator;

/**
 * A utility class providing methods to retrieve configuration information.
 * Initialize by calling <pre> setConfiguration(null). </pre>
 * The configuration location can be specified during startup by using the
 * <BR> java -Dintercommit.configuration=<some location>.
 * Simirarly, a log4j properties file name (but not location) can be specified using
 * <BR> java -Dlog4j.configuration=<some log4j properties file>
 * <BR> Locations can point to files/directories only at the moment, see the TODO.
 * yet been tested.
 * TODO: log4j can only watch a file at the moment, to watch an URL implement the
 * example shown at http://www.laliluna.de/log4j-tutorial.html
 * and then the conf-location can also be an URL.

 * @author frederikw
 *
 */
public final class ConfUtil {

	public static final String fileSeparator = System.getProperty("file.separator");
	public static final String confPropKey = "intercommit.configuration.location";
	public static final String log4jFileNameProp = "log4j.configuration";
	public static final String log4jDefaultFile1 = "log4j.properties";
	/** XML-log file is preferred, it provides more logging options then the properties-file. */
	public static final String log4jDefaultFile2 = "jboss-log4j.xml";
	public static final String jndiDefaultFile = "jndi.properties";
	public static final String jdbcDefaultFile = "jdbc.properties";
	public static final String jmxDefaultFile = "jmxoptions.properties";
	public static final String securityDefaultFile ="adminclient.policy";


	/** The default configuration directory. */
	public static final String defaultConfLocation = "resources/config" + fileSeparator;

	/** The configuration directory. Should always end with a path-separator. */
	public static String confLocation = defaultConfLocation;
	//private static boolean confLocationIsURL = false;
	private static Logger log = null;
	private static final ConfUtil thisInstance = new ConfUtil();

	/** Sets the confLocation according to the sytem property confPropKey,
	 * if it exists. */
	public static void initConfLocation() {

		String loc = System.getProperty(confPropKey);
		if (loc != null) {
			if (loc.length() == 0)
				loc = ".";
			if (!((loc.endsWith("/") || loc.endsWith("\\")))) {
				loc = loc + "/";
			}
			confLocation = loc;
		}
	}

	/** Configures log4j. This method should be called before a logger is
	 * initialized.
	 */
	public static void loadLog4j() {

		/* See also http://logging.apache.org/log4j/1.2/manual.html > Default Initialization Procedure
		 * for the initialization implementation of log4j.
		 */
		System.setProperty("log4j.defaultInitOverride", "true");
		BasicConfigurator.configure();
		if (log == null) {
			log = Logger.getLogger(ConfUtil.class.getName());
		}
		String log4jFileName = System.getProperty(log4jFileNameProp);
		if (log4jFileName == null || !isValidFile(log4jFileName)) {
			if (log4jFileName != null && isValidFile(confLocation + log4jFileName)) {
				log4jFileName = confLocation + log4jFileName;
			}
		}
		if (log4jFileName == null && isValidFile(confLocation + log4jDefaultFile1)) {
			log4jFileName = confLocation + log4jDefaultFile1;
		}
		if (log4jFileName == null && isValidFile(confLocation + log4jDefaultFile2)) {
			log4jFileName = confLocation + log4jDefaultFile2;
		}
		if (log4jFileName != null) {
			openLog(log4jFileName);
			return;
		}
		System.setProperty("log4j.defaultInitOverride", "false");
		Logger.getRootLogger();
		log.warn("Using log4j default configuration.");
	}

	private static void openLog(final String fileNameAndPath) {

		if (log == null) {
			log = Logger.getLogger(ConfUtil.class);
		}
		if (fileNameAndPath.endsWith(".xml")) {
			BasicConfigurator.resetConfiguration();
			DOMConfigurator.configureAndWatch(fileNameAndPath);
			log.info("Log4j configured from: " + fileNameAndPath);
		} else {
			BasicConfigurator.resetConfiguration();
			PropertyConfigurator.configureAndWatch(fileNameAndPath);
			log.info("Log4j configured from: " + fileNameAndPath);
		}
	}

	/** Loads the properties from default JNDI file. If the file cannot be opened,
	 * the properties are set programmatically.
	 * @return The (default) JNDI properties.*/
	public static Properties loadJNDI() {

		Properties props = null;
		props = loadProps(jndiDefaultFile);
		if (props == null) {
			props = new Properties();
			if (log == null) {
				log = Logger.getLogger(ConfUtil.class);
			}
			log.info("Setting jndi.properties programmatically.");
			props.put("java.naming.factory.initial", "org.jnp.interfaces.NamingContextFactory");
			props.put("java.naming.provider.url", "jnp://localhost:1099");
			props.put("java.naming.factory.url.pkgs", "org.jboss.naming:org.jnp.interfaces");
		}
		return props;
	}

	/** Loads the properties from default JMX file and sets them as System properties.
	 * If the file cannot be opened, the properties are set programmatically.
	 * @return The (default) JNDI properties.*/
	public static void loadJMX() {

		Properties props = null;
		props = loadProps(jmxDefaultFile);
		if (props == null) {
			if (log == null) {
				log = Logger.getLogger(ConfUtil.class);
			}
			log.info("Setting jmc.properties programmatically.");
			System.setProperty("com.sun.management.jmxremote.port", "1400");
			System.setProperty("com.sun.management.jmxremote.authenticate", "false");
			System.setProperty("com.sun.management.jmxremote.ssl", "false");
		} else {
			Enumeration<Object> pk = props.keys();
			while (pk.hasMoreElements()) {
				String key = (String)pk.nextElement();
				System.setProperty(key, props.getProperty(key));
			}
		}
	}

	/** Security policy can be required to do a ctx.lookup("x"),
	 * otherwise this will throw an RMI exception.
	 * NOTE: Requires config location with adminclient.policy file with security settings!
	 * RMISecurityManager is set as security manager.
	 * @param securityPolicyFileName The name of the security policy file. If the file-name
	 * does not include a path, the config-location is used. If null,
	 * property java.security.policy is used. If that does not exist,
	 * configLocation + "adminclient.policy" is used.
	 * @return False if the security file could not be found (no security manager is set).
	 */
	public static boolean setSecurity(final String securityPolicyFileName) {

		String spfile = null;
		spfile = fileExists(securityPolicyFileName);
		if (spfile == null)
			spfile = fileExists(confLocation + securityPolicyFileName);
		if (spfile == null)
			spfile = fileExists(System.getProperty("java.security.policy"));
		if (spfile == null)
			spfile = fileExists(confLocation + securityDefaultFile);
		if (log == null) {
			log = Logger.getLogger(ConfUtil.class);
		}
		if (spfile != null) {
			log.info("Using client security config from " + spfile);
			System.setProperty("java.security.policy", spfile);
			System.setSecurityManager(new RMISecurityManager());
		} else {
			log.error("Could not set security manager, security policy file not found.");
		}
		return (spfile != null);
	}

	/**
	 * Loads the properties from the given file from the config-location.
	 * If this fails, tries to open the file via the class-loader.
	 * @param propFileName The name of the file without the config-location path.
	 * @return Null on error, else the loaded properties from the given file.
	 */
	public static Properties loadProps(final String propFileName) {

		Properties props = new Properties();
		boolean propsLoaded = false;
		InputStream is = getFileStream(confLocation + propFileName);
		if (is == null)
			is = getResourceStream(propFileName);
		BufferedInputStream in = null;
		if (log == null) {
			log = Logger.getLogger(ConfUtil.class.getName());
		}
		if (is != null) {
			try {
				in = new BufferedInputStream(is);
				props.load(in);
				propsLoaded = true;
				log.info("Properties loaded from: " + confLocation + propFileName);
			} catch (Exception e) {
				log.error("Error while trying to load properties from:\n" + confLocation + propFileName, e);
			} finally {
				if (in != null) {
					try { in.close(); } catch (IOException unused) {}
				}
			}
		}
		return (propsLoaded) ? props : null;
	}

	/** Tries to open the file. */
	public static InputStream getFileStream(final String fileName) {

		InputStream in = null;
		if (isValidFile(fileName)) {
			try {
				in = new FileInputStream(new File(fileName));
			} catch (FileNotFoundException e) {}
		}
		return in;
	}

	/** Tries to open the resource via the class-loader. */
	public static InputStream getResourceStream(final String fileName) {
		InputStream in = null;
		if (in == null)
			in = ClassLoader.getSystemResourceAsStream(fileName);
		if (in == null)
			in = thisInstance.getClass().getClassLoader().getResourceAsStream(fileName);
		if (in == null)
			in = Thread.currentThread().getContextClassLoader().getResourceAsStream(fileName);
		return in;
	}

	/** Returns true if the file exists. Returns false if the
	 * given filename is a directory.
	 */
	public static boolean isValidFile(final File f) {

		if (f == null)
			return false;
		if (!f.exists())
			return false;
		if (f.isFile())
			return true;
		return false;
	}

	/** Returns true if the file exists. Returns false if the
	 * given filename is a directory.
	 */
	public static boolean isValidFile(final String fileName) {

		if (fileName == null || fileName.length() == 0)
			return false;
		return (isValidFile(new File(fileName)));
	}

	/** Returns null if file does not exist, otherwise the full path of the file. */
	public static String fileExists(final String fileName) {

		if (isValidFile(fileName)) {
			File f = new File(fileName);
			String fcan = null;
			try {
				fcan = f.getCanonicalPath();
			} catch (Exception unused) {}
			return (fcan != null) ? fcan : f.getAbsolutePath();
		}
		return null;
	}

	/** Returns null if the dir does not exist,
	 * otherwise the full path of the dir ending with a file separator. */
	public static String dirExists(final String fileName) {
		if (fileName == null || fileName.length() == 0)
			return null;
		File f = new File(fileName);
		if (!f.exists())
			return null;
		if (f.isDirectory()) {
			String fpath = null;
			try {
				fpath = f.getCanonicalPath();
			} catch (Exception unused) {
				fpath = f.getAbsolutePath();
			}
			if (!fpath.endsWith(fileSeparator))
				fpath = fpath + fileSeparator;
			return fpath;
		}
		return null;
	}
}
