package nl.intercommit.mario.tools;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Properties;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.jboss.security.client.SecurityClient;
import org.jboss.security.client.SecurityClientFactory;

/** Logs-in and retrieves the inital context.
 * After login, lc.logout() must be called before loosing any reference to this object instance.
 * Create using information from: 
 * <br> http://www.jboss.org/index.html?module=bb&op=viewtopic&p=3927618#3927618
 * <br> and 
 * <br> http://www.osintegrators.com/node/24
 * <br>For JSP the SecurityClient must be used and it probably works better in any case, see:
 * <br> http://www.jboss.org/file-access/default/members/jbossejb3/freezone/docs/tutorial/1.0.8/html/index.html (ch. 27 "SecurityClient")
 * <br> http://www.jboss.org/file-access/default/members/jbosssecurity/freezone/releases/guide/2.0.1.GA/jbosssecurity.html
 * <br> http://fisheye.jboss.org/browse/JBossAS/projects/security/security-spi/trunk/
 * @author frederikw
 *
 */
public class MarioContext {
	
	/** The start directory of the application. */
	public static String appHomeDir = System.getProperty("user.dir");
	
	/** The configuration directory containing jndiFile and authFile. */
	public static String confDirName = "conf";
	
	/** The jndi.properties file name. */
	public static String jndiFileName = "jndi.properties";
	
	/** The name of the file containing the JBoss login-module setting. 
	 * <br>See also http://www.jboss.org/community/wiki/ClientLoginModule/ 
	 */
	public static String authFileName = "auth.conf";
	
	public InitialContext ic = null;
	/* Old method
	public LoginContext lc = null; */
	public SecurityClient lc = null;

	/** Set from jndi-property InitialContext.SECURITY_PRINCIPAL. Re-used when reconnecting. */ 
	public String userName = null;
	/** Set from jndi-property InitialContext.SECURITY_CREDENTIALS. Re-used when reconnecting. */
	public String passWord = null;
	
	/** Set when connecting and re-used when reconnecting. */
	public Logger log = null;
	/** Set when connecting and re-used when reconnecting. */
	public boolean login = false;
	/** Set when connecting and re-used when reconnecting. */
	public Hashtable<?,?> icEnv = null;

	/** Does nothing. Connect can be called to initiate the contexts. */
	public MarioContext() {
		super();
	}
	/**
	 * Creates a login-context and initial-context.
	 * Extracts username and password from the jndi-props file.
	 * If username is empty, no login-context is created (in which case initial context uses an anonymous login).
	 * If anything went wrong, an error is logged and ic stays at null.
	 * If log is null, messages are printed using System.out/err
	 */
	public MarioContext(Logger log) {
		this(log, true);
	}

	/**
	 * Creates a login-context (if login is true) and initial-context.
	 * Extracts username and password from the jndi-props file.
	 * If username is empty, no login-context is created (in which case initial context uses an anonymous login).
	 * If anything went wrong, an error is logged and ic stays at null.
	 * If log is null, messages are printed using System.out/err
	 */
	public MarioContext(Logger log, boolean login) {

		this.log = log;
		this.login = login;
		Properties jndiProps = loadJndiProps();
		if (jndiProps != null)
			connect(log, login, jndiProps);
	}
	
	/**
	 * Loads the jndi-properties from jndiFileName.
	 * Username and password are set using the jndi-properties principal and credentials.
	 * @return The loaded properties, null on error.
	 */
	public Properties loadJndiProps() {
		
		FileInputStream fin = null;
		Properties jndiProps = null;
		boolean loadPropsOK = false;
		File propsFile = new File(appHomeDir + File.separator + confDirName + File.separator + jndiFileName);
		try {
			jndiProps = new Properties();
			if (propsFile.exists() && propsFile.isFile()) {
				show("Using file " + propsFile.getAbsolutePath());
				fin = new FileInputStream(appHomeDir + File.separator + confDirName + File.separator + jndiFileName);
				jndiProps.load(fin);
			} else {
				show("Setting JNDI properties programmatically (could not find " + propsFile.getAbsolutePath() + ")");
				jndiProps.put(InitialContext.INITIAL_CONTEXT_FACTORY, "org.jnp.interfaces.NamingContextFactory");
				jndiProps.put(InitialContext.URL_PKG_PREFIXES, "org.jboss.naming:org.jnp.interfaces");
				jndiProps.put(InitialContext.PROVIDER_URL, "jnp://localhost:1099");
			}
			loadPropsOK = true;
			userName = jndiProps.getProperty(InitialContext.SECURITY_PRINCIPAL);
			passWord = jndiProps.getProperty(InitialContext.SECURITY_CREDENTIALS);
		} catch (IOException ioe) {
			if (log != null) {
				log.error("Failed to load jndi-properties required to connect to JBoss server: " + ioe.getMessage(), ioe);
			} else {
				ioe.printStackTrace();
			}
		} finally {
			if (fin != null) try { fin.close(); } catch (Exception ignored) {}
		}
		return loadPropsOK ? jndiProps : null;
	}
	
	/**
	 * Creates a login-context (if login is true) and initial-context.
	 * Username and password have to be already set (they are not extracted from the jndi-props).
	 * If username is empty, no login-context is created (in which case initial context uses an anonymous login).
	 * If anything went wrong, an error is logged and ic stays at null.
	 * If log is null, messages are printed using System.out/err
	 */
	public void connect(Logger log, boolean login, final Properties jndiProps) {
		
		this.log = log;
		this.login = login;
		try {
			login();
			ic = new InitialContext(jndiProps);
			try { icEnv = ic.getEnvironment();} catch (NamingException e) {}
		} catch (Exception e) {
			if (log != null) {
				log.error("Failed to connect to JBoss server: " + e.getMessage(), e);
			} else {
				e.printStackTrace();
			}
			close();
		}
	}
	
	private void login() throws Exception {
		
		if (login && userName != null && userName.length() > 0) {
			/*
			final String authFilePath = appHomeDir + File.separator + confDirName + File.separator + authFileName; 
			show("Using file " + authFilePath); 
			System.setProperty("java.security.auth.login.config", authFilePath);
			lc = new LoginContext("MarioLogin", new UsernamePasswordHandler(userName, passWord));
			lc.login();
			*/
			lc = SecurityClientFactory.getSecurityClient();
			lc.setVmwideAssociation(true);
			lc.setSimple(userName, passWord);
			lc.login();
		} else if (login) {
			show("Could not find a username, using anonymous login.");
		}
	}
	
	/** Attempts to reconnect using data from previous connect. If it fails, ic is null. */
	public void reconnect() {
		close();
		try {
			login();
			ic = new InitialContext(icEnv);
		} catch (Exception e) {
			if (log != null) {
				log.error("Failed to connect to JBoss server: " + e.getMessage(), e);
			} else {
				e.printStackTrace();
			}
			close();
		}
	}
	
	/** Logs out, closes the contexts and sets lc and ic to null. */
	public void close() {
		
		if (lc != null) try { lc.logout(); } catch (Exception unused) {}
		lc = null;
		if (ic != null) try { ic.close(); } catch (Exception unused) {}
		ic = null;
	}
	
	private void show(final String infoText) {

		if (log != null) {
			log.info(infoText);
		} else {
			System.out.println(infoText);
		}
	}
	
}
